
from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesLabel, CssStylesDates, CssStylesInput


class CatalogInput(Catalog.CatalogGroup):
  def active(self):
    """

    """
    return self._set_class(CssStylesInput.CssUIActive)

  def basic(self):
    """
    Description:
    -----------
    Basic style for an input component
    """
    return self._set_class(CssStylesInput.CssInput)

  def range(self):
    """
    Description:
    -----------
    CSS Style for the input range component
    """
    return self._set_class(CssStylesInput.CssInputRange)

  def range_thumb(self):
    """
    Description:
    -----------
    CSS Style for the thumb of the input range component
    """
    return self._set_class(CssStylesInput.CssInputRangeThumb)

  def label(self):
    """
    Description:
    -----------
    CSS Style for the label attached to an input component
    """
    return self._set_class(CssStylesInput.CssInputLabel)

  def label_hover(self):
    """
    Description:
    -----------
    CSS Style to change the label background color on mouse hover
    """
    return self._set_class(CssStylesLabel.CssLabelCheckMarkHover)

  def label_disable(self):
    """
    Description:
    -----------
    CSS Style to set the label to be disabled
    """
    return self._set_class(CssStylesLabel.CssLabelContainerDisabled)

  def integer(self):
    """
    Description:
    -----------
    Basic style for an input integer component
    """
    return self._set_class(CssStylesInput.CssInputInteger)

  def text(self):
    """
    Description:
    -----------
    Basic style for an input text component
    """
    return self._set_class(CssStylesInput.CssInputText)

  def textarea(self):
    """
    Description:
    -----------
    Basic style for an input textarea component
    """
    return self._set_class(CssStylesInput.CssInputTextArea)

  def is_valid(self):
    """
    Description:
    -----------
    Basic style for an input component with a valid condition
    """
    return self._set_class(CssStylesInput.CssInputValid)


class CatalogDate(Catalog.CatalogGroup):
  def datepicker_ui(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesDates.CssDatePickerUI)

  def datepicker(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesDates.CssDatePicker)

  def datepicker_header(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesDates.CssDatesDatePickerHeader)

  def time_picker(self):
    """
    Description:
    ------------
    CSS class in charge of changing the container of the different possible items to be selected
    """
    return self._set_class(CssStylesDates.CssDatesTimePicker)

  def time_picker_items(self):
    """
    Description:
    ------------
    CSS class in charge of changing the color of the items on mouse hover.
    This will by default change the background of the item
    """
    return self._set_class(CssStylesDates.CssDatesTimePickerState)




from epyk.core.js.html import JsHtml
from epyk.core.js.fncs import JsFncs
from epyk.core.js import JsUtils

from epyk.core.js.primitives import JsObjects


class JsHtmlDatePicker(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.val(), timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
      self.htmlCode, self._src.dom.jquery.varId))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, '%s.val()' % self._src.dom.jquery.varId)


class JsHtmlDateFieldPicker(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.val, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
      self.htmlCode, self.content.toStr()))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, '%s.val()' % self._src.input.dom.jquery.varId)


class JsHtmlProgressBar(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.progressbar('value'), timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self._src.dom.jquery.varId))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, '%s.progressbar("value")' % self._src.dom.jquery.varId)

  def position(self, val, jsFnc):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param val:
    :param jsFnc:
    """
    return JsFncs.JsFunction("if((%(content)s >= %(val)s) && (%(content)s <= %(val)s + 10)){%(fnc)s}" % {'content': self.content, 'val': val, 'fnc': JsUtils.jsConvertFncs(jsFnc, toStr=True)})

  def max(self, jsFnc):
    """
    Description:
    ------------

    :param jsFnc:
    """
    return self.position(100, jsFnc)


class JsHtmlTimePicker(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.val(), timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self._src.dom.jquery.varId))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, '%s.val()' % self._src.dom.jquery.varId)


class JsHtmlSlider(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.slider('value'), timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self._src.dom.jquery.varId))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, '%s.slider("value")' % self._src.dom.jquery.varId)


class JsHtmlSliderRange(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.slider('values'), timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self._src.dom.jquery.varId))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, '%s.slider("values")' % self._src.dom.jquery.varId)


class JsHtmlSliderDate(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self.content.toStr()))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, 'new Date(%s.slider("value") * 1000).toISOString().split("T")[0]' % self._src.dom.jquery.varId)


class JsHtmlSliderDates(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self.content.toStr()))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    return JsHtml.ContentFormatters(self._report, 'function() {return [new Date(%s.slider("values")[0] * 1000).toISOString().split("T")[0], new Date(%s.slider("values")[1] * 1000).toISOString().split("T")[0]]}()' % (self._src.dom.jquery.varId, self._src.dom.jquery.varId))


class JsHtmlSparkline(JsHtml.JsHtml):

  @property
  def val(self):
    """
    Description:
    ------------
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (self.htmlCode, self.region))

  @property
  def content(self):
    """
    Description:
    ------------

    """
    if self._src._jsStyles['type'] in ['bar']:
      return JsHtml.ContentFormatters(self._report, 'event.sparklines[0].getCurrentRegionFields()[0].value')

    return JsHtml.ContentFormatters(self._report, 'event.sparklines[0].getCurrentRegionFields().y')

  @property
  def value(self):
    if self._src._jsStyles['type'] in ['bar']:
      return JsHtml.ContentFormatters(self._report, 'event.sparklines[0].getCurrentRegionFields()[0].value')

    return JsObjects.JsNumber.JsNumber("event.sparklines[0].getCurrentRegionFields().y", isPyData=False)

  @property
  def region(self):
    return JsObjects.JsObjects.get("event.sparklines[0].getCurrentRegionFields()")

  @property
  def x(self):
    if self._src._jsStyles['type'] in ['bar']:
      return JsObjects.JsObjects.get("event.sparklines[0].getCurrentRegionFields().offset")

    return JsObjects.JsObjects.get("event.sparklines[0].getCurrentRegionFields().x")

  @property
  def offset(self):
    if self._src._jsStyles['type'] in ['bar']:
      return JsObjects.JsObjects.get("event.sparklines[0].getCurrentRegionFields()[0].offset")

    return JsObjects.JsObjects.get("event.sparklines[0].getCurrentRegionFields().offset")

  @property
  def y(self):
    if self._src._jsStyles['type'] in ['bar']:
      return JsObjects.JsObjects.get("event.sparklines[0].getCurrentRegionFields()[0].value")

    return JsObjects.JsNumber.JsNumber("event.sparklines[0].getCurrentRegionFields().y", isPyData=False)



"""

"""

from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesHr, CssStylesScrollBar
from epyk.core.css.styles.classes import CssStylesSearch, CssStylesCommon, CssStylesDivMenuBars, CssStylesPanel


class CatalogLayout(Catalog.CatalogGroup):
  def hr(self):
    """  """
    return self._set_class(CssStylesHr.CssHr)

  def body(self):
    """  """
    return self._set_class(CssStylesCommon.CssBody)

  def body_content(self):
    """  """
    return self._set_class(CssStylesCommon.CssBodyContent)

  def loading(self):
    """  """
    return self._set_class(CssStylesCommon.CssBodyLoadingBack)

  def selection(self):
    """  """
    return self._set_class(CssStylesCommon.CssTextSelection)

  def panel_arrow_down(self):
    """  """
    return self._set_class(CssStylesPanel.CssPanelArrowDown)

  def panel_arrow_up(self):
    """  """
    return self._set_class(CssStylesPanel.CssPanelArrowUp)

  def menubar(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssSideBarMenu)

  def menubar_fixed(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssSideBarFixed)

  def menubar_bubble(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssSideBarBubble)

  def menubar_side(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssSideBar)

  def menubar_link(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssSideBarLiHref)

  def menubar_item(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssSideBarLi)

  def menubar_params(self):
    """  """
    return self._set_class(CssStylesDivMenuBars.CssParamsBar)

  def search(self):
    """  """
    return self._set_class(CssStylesSearch.CssSearch)

  def search_extension(self):
    """  """
    return self._set_class(CssStylesSearch.CssSearchExt)

  def search_button(self):
    """  """
    return self._set_class(CssStylesSearch.CssSearchButton)

  def scrollbar(self):
    """  """
    return self._set_class(CssStylesScrollBar.CssScrollBar)

  def scrollbar_track(self):
    """  """
    return self._set_class(CssStylesScrollBar.CssScrollBarTrack)

  def scrollbar_track_thumb(self):
    """  """
    return self._set_class(CssStylesScrollBar.CssScrollBarTrackThumb)



"""

TODO: add the is_secured logic in this module
TODO: Check Neo4J
"""

import os

from epyk.core.js.Imports import requires
from epyk.core.py import PySql


class NoSql(object):
  class __internal(object):
    _props = {}

  def __init__(self, report=None):
    self._report = report if report is not None else self.__internal()

  def mongo(self, host="localhost", port=5000, is_secured=False):
    """

    Related Pages:

			https://www.w3schools.com/python/python_mongodb_getstarted.asp
    https://www.mongodb.com/dr/fastdl.mongodb.org/win32/mongodb-win32-x86_64-2008plus-ssl-4.0.3-signed.msi/download

    :param host: Optional, Database hostname. Default localhost
    :param port: Optional, Database port. Default 5000
    :param is_secured:

    :return:

    """
    pyMongo = requires("pyMongo", reason='Missing Package', install="pyMongo", source_script=__file__, raise_sxcept=True)
    return pyMongo.MongoClient("mongodb://%s:%s/" % (host, port))

  def neo4j(self, host="localhost", port=5000, is_secured=False):
    """

    Related Pages:

			https://neo4j.com/developer/python/
    https://community.neo4j.com/?_ga=2.69585106.394662973.1539480121-615400289.1539480121

    :param host: Optional, Database hostname. Default localhost
    :param port: Optional, Database port. Default 5000
    :param is_secured:

    :return: A Python SQL connectionr for Neo4J
    """
    if 'neo4j' not in self.pkgs:
      requires("neo4j", reason='Missing Package', install='neo4j-driver', source_script=__file__, raise_sxcept=True)
    return PySql.SqlConnNeo4j(host, port)


class DataDb(object):
  class __internal(object):
    _props = {}

  def __init__(self, report=None):
    self._report = report if report is not None else self.__internal()
    self._db_bindings, self.pkgs = {}, {}
    self.no_sql = NoSql(report)
    self.table_names = None # create the unique set of tables to be loaded

  def __settings(self):
    """
    Retrieve the database settings based on the environment configuration.

    Settings are used in the context of generic private and public databases.
    If SQLite is used, the database will be automatically generated.

    :return: The database settings
    """
    db_settings = {"family": 'sqlite'} if self._report._dbSettings is None else dict(self._report._dbSettings)
    if not "family" in db_settings:
      db_settings["family"] = 'sqlite'
    if 'names' in db_settings:
      db_settings['public'] = db_settings['names']['public']['name']
      if db_settings['names']['private'].get("fixed", False):
        db_settings['private'] = db_settings['names']['private']['name']
      else:
        db_settings['private'] = "%s_%s" % (db_settings['names']['private']['name'], self._report.py.hash(self._report.run.current_user))
    elif 'name' in db_settings:
      # In this context there is no unique database per user and all the users will use a different
      # DB when they are in a private mode
      db_settings['private'] = db_settings['name']
      db_settings['public'] = db_settings['name']
      del db_settings['name']

    else:
      if not 'path' in db_settings:
        db_settings['path'] = self._report.run.local_path
      db_settings['private'] = '%s_%s' % (self._report.run.report_name, self._report.py.hash(self._report.run.current_user))
      db_settings['public'] = self._report.run.report_name
    if not 'model_path' in db_settings:
      db_settings['model_path'] = os.path.join(self._report.run.local_path, 'model', 'tables')
    return db_settings

  @property
  def names(self):
    """
    Returns the different database names to be used by the user in a public or private context.

    Generally the public database is shared by all the users whereas the private one is only dedicated to a user and
    data can be granted to some external users.

    Example

    Related Pages:

			https://www.sqlalchemy.org/

    :return: A dictionary with the database names in each context
    """
    settings = self.__settings()
    return {"public": settings['public'], "private": settings['private']}

  @property
  def private(self):
    """
    Return the private database from the family defined in the environment.

    It is not possible to specify the type of DB for an environment at this level.
    This should be done either in the server configuration or in the report configuration.

    This module rely on SQLAlchemy for the query generation

    Example

    Related Pages:

			https://www.sqlalchemy.org/

    :rtype: epyk.core.py.PySql.SqlConn
    :return:
    """
    settings = self.__settings()
    database = "%(path)s/%(private)s" % settings if settings.get("path") is not None else settings['private']
    if database not in self._db_bindings:
      model_path = os.path.join(self._report.run.local_path, 'model', 'tables')
      if settings["family"] == "sqlite":
        db = self.sqlite(settings["private"], settings.get("path"), model_path, tables_scope=self.table_names)
      elif settings["family"] == "mssql":
        db = self.mssql(settings["private"], settings["host"], driverName=settings["driverName"], model_path=model_path, tables_scope=self.table_names)
      elif settings["family"] == "oracle":
        db = self.oracle(settings["private"], settings["host"], settings["port"], model_path=model_path, tables_scope=self.table_names)
      elif settings["family"] == "postgres":
        db = self.postgres(settings["private"], settings["host"], settings["port"], model_path=model_path, tables_scope=self.table_names)
      elif settings["family"] == "mysql":
        db = self.mysql(settings["private"], settings["host"], settings["port"], model_path=model_path, tables_scope=self.table_names)
      elif settings["family"] == "mariadb":
        db = self.mariadb(settings["private"], settings["host"], settings["port"], model_path=model_path, tables_scope=self.table_names)
      elif settings["family"] == "mdb":
        db = self.mdb(settings["private"], settings.get("path"), model_path)
      elif settings["family"] == "accdb":
        db = self.accdb(settings["private"], settings.get("path"), model_path)
      self._db_bindings[database] = db
    return self._db_bindings[database]

  @property
  def public(self):
    """
    Return the public shared database from the family defined in the environment.

    It is not possible to specify the type of DB for an environment at this level.
    This should be done either in the server configuration or in the report configuration.

    This module rely on SQLAlchemy for the query generation

    Example

    Related Pages:

			https://www.sqlalchemy.org/

    :rtype: epyk.core.py.PySql.SqlConn

    :return: A Sql object
    """
    settings = self.__settings()
    database = "%(path)s/%(public)s" % settings if settings.get("path") is not None else settings['public']
    if database not in self._db_bindings:
      model_path = os.path.join(self._report.run.local_path, 'model', 'tables')
      if settings["family"] == "sqlite":
        db = self.sqlite(settings["public"], settings.get("path"), model_path)
      elif settings["family"] == "mssql":
        db = self.mssql(settings["public"], settings["host"], driverName=settings["driverName"], model_path=model_path)
      elif settings["family"] == "oracle":
        db = self.oracle(settings["public"], settings["host"], settings["port"], model_path=model_path)
      elif settings["family"] == "postgres":
        db = self.postgres(settings["public"], settings["host"], settings["port"], model_path=model_path)
      elif settings["family"] == "mysql":
        db = self.mysql(settings["public"], settings["host"], settings["port"], model_path=model_path)
      elif settings["family"] == "mariadb":
        db = self.mariadb(settings["public"], settings["host"], settings["port"], model_path=model_path)
      elif settings["family"] == "mdb":
        db = self.mdb(settings["public"], settings.get("path"), model_path)
      elif settings["family"] == "accdb":
        db = self.accdb(settings["public"], settings.get("path"), model_path)
      self._db_bindings[database] = db
    return self._db_bindings[database]

  def reflect(self, table_names):
    """
    Reduce the scope of database tables to be loaded by SQLAchemy

    Related Pages:

			https://docs.sqlalchemy.org/en/13/core/metadata.html

    :param table_names:

    :return:
    """
    if self.table_names is None:
      self.table_names = set(table_names)
    else:
      self.table_names |= set(table_names)
    return self

  # -------------------------------------------------
  # Specific database entries
  #
  def sqlite(self, name, db_path=None, model_path=None, db_settings=None, tables_scope=None):
    """
    Create a bespoke SQLite database

    By default the database will be in the current folder.
    The table definition will not be done automatically for this database and this should be created in a external manner

    Example

    Documentation

    :param name: The database name
    :param db_path: Optional, the database path
    :param model_path: Optional, the model with the tables definition. The filename must be unique in the project

    :rtype: epyk.core.py.PySql.SqlConn

    :return: The SQL object
    """
    if db_path is None:
      db_path = self._report.run.local_path
    database = "%s/%s.db" % (db_path, name)
    dataSettings = {"loadModel": model_path is not None, 'model_path': model_path or False}
    if db_settings is not None:
      dataSettings.update(db_settings)
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConn('sqlite', database=database, tables_scope=tables_scope, **dataSettings)
    return self._db_bindings[database]

  def oracle(self, name, host, port, model_path=None, is_secured=False, tables_scope=None):
    """

    Example

    Documentation

    :param name: Database name
    :param host: Optional, Database hostname. Default localhost
    :param port: Optional, Database port. Default 5432
    :param model_path: Optional, Database model path with the python scripts of the tables
    :param is_secured: If credentials required. Default False

    :rtype: epyk.core.py.PySql.SqlConn

    :return:
    """
    if 'oracle' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="cx_Oracle", source_script=__file__)

    database = name
    db_settings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                   "username": "postgres", "password": "240985", "host": host, "port": port}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConn('oracle', database=database, tables_scope=tables_scope, **db_settings)
    return self._db_bindings[database]

  def mssql(self, name, host="localhost", driverName="{ODBC Driver 17 for SQL Server}", model_path=None, tables_scope=None):
    """

    Example

    Documentation

    :param name: The database name
    :param host: The database host name
    :param driverName: Optional, The
    :param model_path: Optinal, The databse model path

    :rtype: epyk.core.py.PySql.SqlConn

    :return:
    """
    if 'mssql' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="mssql", source_script=__file__, raise_except=True)
    if 'pyodbc' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="pyodbc", source_script=__file__, raise_except=True)
    database = name
    dataSettings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                    "driver": 'mssql+pyodbc', "driverName": driverName, 'host': host}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConn('mssql+pyodbc', database=database, tables_scope=tables_scope, **dataSettings)
    return self._db_bindings[database]

  def mdb(self, name, db_path, model_path=None):
    """
    Get a Access (.mdb) Database query object using ODBC driver

    Example

    Documentation

    :param name: The filename
    :param db_path: The database full path
    :param model_path: Optional, Database model path with the python scripts of the tables

    :rtype: epyk.core.py.PySql.SqlConnOdbc

    :return:
    """
    if 'pyodbc' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="pyodbc", source_script=__file__)
    database = "%s/%s.mdb" % (db_path, name)
    dataSettings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                    "driver": "{Microsoft Access Driver (*.mdb, *.accdb)}"}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConnOdbc(database=database, **dataSettings)
    return self._db_bindings[database]

  def accdb(self, name, db_path, model_path=None):
    """
    Get a Access (.accdb) Database query object using ODBC driver

    Example

    Related Pages:

			https://www.599cd.com/access/studentdatabases/

    :param name: The filename
    :param db_path: The database full path
    :param model_path: Optional, Database model path with the python scripts of the tables

    :rtype: epyk.core.py.PySql.SqlConnOdbc

    :return:
    """
    if 'pyodbc' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="pyodbc", source_script=__file__)
    database = "%s/%s.accdb" % (db_path, name)
    dataSettings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                    "driver": "{Microsoft Access Driver (*.mdb, *.accdb)}"}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConnOdbc(database=database, **dataSettings)
    return self._db_bindings[database]

  def postgres(self, name, host="localhost", port=5432, model_path=None, is_secured=False, tables_scope=None):
    """
    Get a PostgreSql Database query object using SQLAlchemy

    Related Pages:

			https://www.postgresql.org/
    https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#sqlalchemy.dialects.postgresql.dml.Insert.on_conflict_do_update.params.where

    :param name: Database name
    :param host: Optional, Database hostname. Default localhost
    :param port: Optional, Database port. Default 5432
    :param model_path: Optional, Database model path with the python scripts of the tables
    :param is_secured: If credentials required. Default False

    :rtype: epyk.core.py.PySql.SqlConn

    :return:
    """
    if 'postgresql' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="postgresql", source_script=__file__, raise_except=True)
    if 'psycopg2' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="psycopg2", source_script=__file__, raise_except=True)
    database = name
    db_settings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                    "username": "postgres", "password": "240985", "host": host, "port": port}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConn('postgresql+psycopg2', database=database, tables_scope=tables_scope, **db_settings)
    return self._db_bindings[database]

  def mysql(self, name, host="localhost", port=3306, model_path=None, is_secured=False, tables_scope=None):
    """
    Get a MySql Database query object using SQLAlchemy

    Example
    rptObj.data.db.mysql("MySQL", port=3333)

    Related Pages:

			https://dev.mysql.com/

    :param name: Database name
    :param host: Optional, Database hostname. Default localhost
    :param port: Optional, Database port. Default 3306
    :param model_path: Optional, Database model path with the python scripts of the tables
    :param is_secured: If credentials required. Default False

    :rtype: epyk.core.py.PySql.SqlConn

    :return:
    """
    if 'pymysql' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="pymysql", source_script=__file__, raise_except=True)
    database = name
    db_settings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                    "username": "root", "password": "240985", "host": host, "port": port}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConn('mysql+pymysql', database=database, tables_scope=tables_scope, **db_settings)
    return self._db_bindings[database]

  def mariadb(self, name, host="localhost", port=3306, model_path=None, is_secured=False, tables_scope=None):
    """
    Get a MariaDB Database query object using SQLAlchemy

    Example
    rptObj.data.db.mariadb("MySQL", port=3333)

    Related Pages:

			https://mariadb.org/

    :param name: Database name
    :param host: Optional, Database hostname. Default localhost
    :param port: Optional, Database port. Default 3306
    :param model_path: Optional, Database model path with the python scripts of the tables
    :param is_secured: If credentials required. Default False

    :rtype: epyk.core.py.PySql.SqlConn
    """
    if 'pymysql' not in self.pkgs:
      self.pkgs[name] = requires(name, reason='Missing Package', install="pymysql", source_script=__file__, raise_except=True)
    database = name
    db_settings = {"loadModel": model_path is not None, 'model_path': model_path or False,
                    "username": "root", "password": "240985", "host": host, "port": port}
    if database not in self._db_bindings:
      self._db_bindings[database] = PySql.SqlConn('mysql+pymysql', database=database, tables_scope=tables_scope, **db_settings)
    return self._db_bindings[database]





from epyk.core.html import Html
from epyk.core.js import JsUtils


class Table(Html.Html):
  name = 'Google Table'
  requirements = ('google-tables', )

  def __init__(self, report, records, width, height, htmlCode, options, profile):
    data, columns, self.__config = [], [], None
    super(Table, self).__init__(report, records, htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.__options = options

  @property
  def tableId(self):
    """
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  def add_column(self, c):
    pass

  def build(self, data=None, options=None, profile=False):
    return '''
      %(chartId)s = google.charts.setOnLoadCallback( (function(){
        var data = new google.visualization.DataTable();
        var tableData = %(data)s;
        tableData.rows.forEach(function(c){
          data.addColumn('string', c)});
        tableData.cols.forEach(function(c){
          data.addColumn('number', c)});
        data.addRows(tableData.datasets);

        var chart = new google.visualization.%(type)s(%(varId)s);
        chart.draw(data, {});
        return chart
      }));
      ''' % {'chartId': self.tableId, 'varId': self.dom.varId, 'data': JsUtils.jsConvertData(data, None),
             'type': self.__options["type"]}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())




from epyk.core.data.DataClass import DataClass

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class ZoomRange(DataClass):

  @property
  def x(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["x"]

  @x.setter
  def x(self, num):
    self._attrs["x"] = num

  @property
  def y(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["y"]

  @y.setter
  def y(self, num):
    self._attrs["y"] = num


class ZoomAttrs(DataClass):

  @property
  def enabled(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, bool):
    self._attrs["enabled"] = bool

  @property
  def mode(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["mode"]

  @mode.setter
  def mode(self, value):
    self._attrs["mode"] = value

  @property
  def rangeMin(self):
    return self.sub_data("rangeMin", ZoomRange)

  @property
  def rangeMax(self):
    return self.sub_data("rangeMax", ZoomRange)

  @property
  def speed(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["speed"]

  @speed.setter
  def speed(self, num):
    self._attrs["speed"] = num

  @property
  def threshold(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["threshold"]

  @threshold.setter
  def threshold(self, num):
    self._attrs["threshold"] = num


class ZoomPan(ZoomAttrs):

  def onPan(self, jsFncs):
    """
    Description:
    -----------
    Function called while the user is zooming

    :param jsFnc:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onPan"] = JsObjects.JsVoid("function(data) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))

  def onPanComplete(self, jsFncs):
    """
    Description:
    -----------
    Function called while the user is zooming

    :param jsFnc:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onPanComplete"] = JsObjects.JsVoid("function(data) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))


class ZoomZoom(ZoomAttrs):

  @property
  def drag(self):
    """
    Description:
    -----------
    Enable drag-to-zoom behavior
    """
    return self._attrs["drag"]

  @drag.setter
  def drag(self, bool):
    self._attrs["drag"] = bool

  @property
  def sensitivity(self):
    """
    Related Pages:

			https://github.com/chartjs/chartjs-plugin-zoom
    """
    return self._attrs["sensitivity"]

  @sensitivity.setter
  def sensitivity(self, num):
    self._attrs["sensitivity"] = num

  def onZoom(self, jsFncs):
    """
    Function called while the user is zooming

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onZoom"] = JsObjects.JsVoid("function(data) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))

  def onZoomComplete(self, jsFncs):
    """
    Function called once zooming is completed

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onZoomComplete"] = JsObjects.JsVoid("function(data) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))


class Zoom(DataClass):

  @property
  def pan(self):
    return self.sub_data("pan", ZoomPan)

  @property
  def zoom(self):
    return self.sub_data("zoom", ZoomZoom)




"""
Mapping table for the pre defined HTML colors

Related Pages:

			https://www.rapidtables.com/web/color/RGB_Color.html
"""

import random
import math


defined = {
  'MAROON': {'hex': '#800000', 'rgb': '(128,0,0)'},
  'DARK RED': {'hex': '#8B0000', 'rgb': '(139,0,0)'},
  'BROWN': {'hex': '#A52A2A', 'rgb': '(165,42,42)'},
  'FIREBRICK': {'hex': '#B22222', 'rgb': '(178,34,34)'},
  'CRIMSON': {'hex': '#DC143C', 'rgb': '(220,20,60)'},
  'RED': {'hex': '#FF0000', 'rgb': '(255,0,0)'},
  'TOMATO': {'hex': '#FF6347', 'rgb': '(255,99,71)'},
  'CORAL': {'hex': '#FF7F50', 'rgb': '(255,127,80)'},
  'INDIAN RED': {'hex': '#CD5C5C', 'rgb': '(205,92,92)'},
  'LIGHT CORAL': {'hex': '#F08080', 'rgb': '(240,128,128)'},
  'DARK SALMON': {'hex': '#E9967A', 'rgb': '(233,150,122)'},
  'SALMON': {'hex': '#FA8072', 'rgb': '(250,128,114)'},
  'LIGHT SALMON': {'hex': '#FFA07A', 'rgb': '(255,160,122)'},
  'ORANGE RED': {'hex': '#FF4500', 'rgb': '(255,69,0)'},
  'DARK ORANGE': {'hex': '#FF8C00', 'rgb': '(255,140,0)'},
  'ORANGE': {'hex': '#FFA500', 'rgb': '(255,165,0)'},
  'GOLD': {'hex': '#FFD700', 'rgb': '(255,215,0)'},
  'DARK GOLDEN ROD': {'hex': '#B8860B', 'rgb': '(184,134,11)'},
  'GOLDEN ROD': {'hex': '#DAA520', 'rgb': '(218,165,32)'},
  'PALE GOLDEN ROD': {'hex': '#EEE8AA', 'rgb': '(238,232,170)'},
  'DARK KHAKI': {'hex': '#BDB76B', 'rgb': '(189,183,107)'},
  'KHAKI': {'hex': '#F0E68C', 'rgb': '(240,230,140)'},
  'OLIVE': {'hex': '#808000', 'rgb': '(128,128,0)'},
  'YELLOW': {'hex': '#FFFF00', 'rgb': '(255,255,0)'},
  'YELLOW GREEN': {'hex': '#9ACD32', 'rgb': '(154,205,50)'},
  'DARK OLIVE GREEN': {'hex': '#556B2F', 'rgb': '(85,107,47)'},
  'OLIVE DRAB': {'hex': '#6B8E23', 'rgb': '(107,142,35)'},
  'LAWN GREEN': {'hex': '#7CFC00', 'rgb': '(124,252,0)'},
  'CHART REUSE': {'hex': '#7FFF00', 'rgb': '(127,255,0)'},
  'GREEN YELLOW': {'hex': '#ADFF2F', 'rgb': '(173,255,47)'},
  'DARK GREEN': {'hex': '#006400', 'rgb': '(0,100,0)'},
  'GREEN': {'hex': '#008000', 'rgb': '(0,128,0)'},
  'FOREST GREEN': {'hex': '#228B22', 'rgb': '(34,139,34)'},
  'LIME': {'hex': '#00FF00', 'rgb': '(0,255,0)'},
  'LIME GREEN': {'hex': '#32CD32', 'rgb': '(50,205,50)'},
  'LIGHT GREEN': {'hex': '#90EE90', 'rgb': '(144,238,144)'},
  'PALE GREEN': {'hex': '#98FB98', 'rgb': '(152,251,152)'},
  'DARK SEA GREEN': {'hex': '#8FBC8F', 'rgb': '(143,188,143)'},
  'MEDIUM SPRING GREEN': {'hex': '#00FA9A', 'rgb': '(0,250,154)'},
  'SPRING GREEN': {'hex': '#00FF7F', 'rgb': '(0,255,127)'},
  'SEA GREEN': {'hex': '#2E8B57', 'rgb': '(46,139,87)'},
  'MEDIUM AQUA MARINE': {'hex': '#66CDAA', 'rgb': '(102,205,170)'},
  'MEDIUM SEA GREEN': {'hex': '#3CB371', 'rgb': '(60,179,113)'},
  'LIGHT SEA GREEN': {'hex': '#20B2AA', 'rgb': '(32,178,170)'},
  'DARK SLATE GRAY': {'hex': '#2F4F4F', 'rgb': '(47,79,79)'},
  'TEAL': {'hex': '#008080', 'rgb': '(0,128,128)'},
  'DARK CYAN': {'hex': '#008B8B', 'rgb': '(0,139,139)'},
  'AQUA': {'hex': '#00FFFF', 'rgb': '(0,255,255)'},
  'CYAN': {'hex': '#00FFFF', 'rgb': '(0,255,255)'},
  'LIGHT CYAN': {'hex': '#E0FFFF', 'rgb': '(224,255,255)'},
  'DARK TURQUOISE': {'hex': '#00CED1', 'rgb': '(0,206,209)'},
  'TURQUOISE': {'hex': '#40E0D0', 'rgb': '(64,224,208)'},
  'MEDIUM TURQUOISE': {'hex': '#48D1CC', 'rgb': '(72,209,204)'},
  'PALE TURQUOISE': {'hex': '#AFEEEE', 'rgb': '(175,238,238)'},
  'AQUA MARINE': {'hex': '#7FFFD4', 'rgb': '(127,255,212)'},
  'POWDER BLUE': {'hex': '#B0E0E6', 'rgb': '(176,224,230)'},
  'CADET BLUE': {'hex': '#5F9EA0', 'rgb': '(95,158,160)'},
  'STEEL BLUE': {'hex': '#4682B4', 'rgb': '(70,130,180)'},
  'CORN FLOWER BLUE': {'hex': '#6495ED', 'rgb': '(100,149,237)'},
  'DEEP SKY BLUE': {'hex': '#00BFFF', 'rgb': '(0,191,255)'},
  'DODGER BLUE': {'hex': '#1E90FF', 'rgb': '(30,144,255)'},
  'LIGHT BLUE': {'hex': '#ADD8E6', 'rgb': '(173,216,230)'},
  'SKY BLUE': {'hex': '#87CEEB', 'rgb': '(135,206,235)'},
  'LIGHT SKY BLUE': {'hex': '#87CEFA', 'rgb': '(135,206,250)'},
  'MIDNIGHT BLUE': {'hex': '#191970', 'rgb': '(25,25,112)'},
  'NAVY': {'hex': '#000080', 'rgb': '(0,0,128)'},
  'DARK BLUE': {'hex': '#00008B', 'rgb': '(0,0,139)'},
  'MEDIUM BLUE': {'hex': '#0000CD', 'rgb': '(0,0,205)'},
  'BLUE': {'hex': '#0000FF', 'rgb': '(0,0,255)'},
  'ROYAL BLUE': {'hex': '#4169E1', 'rgb': '(65,105,225)'},
  'BLUE VIOLET': {'hex': '#8A2BE2', 'rgb': '(138,43,226)'},
  'INDIGO': {'hex': '#4B0082', 'rgb': '(75,0,130)'},
  'DARK SLATE BLUE': {'hex': '#483D8B', 'rgb': '(72,61,139)'},
  'SLATE BLUE': {'hex': '#6A5ACD', 'rgb': '(106,90,205)'},
  'MEDIUM SLATE BLUE': {'hex': '#7B68EE', 'rgb': '(123,104,238)'},
  'MEDIUM PURPLE': {'hex': '#9370DB', 'rgb': '(147,112,219)'},
  'DARK MAGENTA': {'hex': '#8B008B', 'rgb': '(139,0,139)'},
  'DARK VIOLET': {'hex': '#9400D3', 'rgb': '(148,0,211)'},
  'DARK ORCHID': {'hex': '#9932CC', 'rgb': '(153,50,204)'},
  'MEDIUM ORCHID': {'hex': '#BA55D3', 'rgb': '(186,85,211)'},
  'PURPLE': {'hex': '#800080', 'rgb': '(128,0,128)'},
  'THISTLE': {'hex': '#D8BFD8', 'rgb': '(216,191,216)'},
  'PLUM': {'hex': '#DDA0DD', 'rgb': '(221,160,221)'},
  'VIOLET': {'hex': '#EE82EE', 'rgb': '(238,130,238)'},
  'MAGENTA / FUCHSIA': {'hex': '#FF00FF', 'rgb': '(255,0,255)'},
  'ORCHID': {'hex': '#DA70D6', 'rgb': '(218,112,214)'},
  'MEDIUM VIOLET RED': {'hex': '#C71585', 'rgb': '(199,21,133)'},
  'PALE VIOLET RED': {'hex': '#DB7093', 'rgb': '(219,112,147)'},
  'DEEP PINK': {'hex': '#FF1493', 'rgb': '(255,20,147)'},
  'HOT PINK': {'hex': '#FF69B4', 'rgb': '(255,105,180)'},
  'LIGHT PINK': {'hex': '#FFB6C1', 'rgb': '(255,182,193)'},
  'PINK': {'hex': '#FFC0CB', 'rgb': '(255,192,203)'},
  'ANTIQUE WHITE': {'hex': '#FAEBD7', 'rgb': '(250,235,215)'},
  'BEIGE': {'hex': '#F5F5DC', 'rgb': '(245,245,220)'},
  'BISQUE': {'hex': '#FFE4C4', 'rgb': '(255,228,196)'},
  'BLANCHED ALMOND': {'hex': '#FFEBCD', 'rgb': '(255,235,205)'},
  'WHEAT': {'hex': '#F5DEB3', 'rgb': '(245,222,179)'},
  'CORN SILK': {'hex': '#FFF8DC', 'rgb': '(255,248,220)'},
  'LEMON CHIFFON': {'hex': '#FFFACD', 'rgb': '(255,250,205)'},
  'LIGHT GOLDEN ROD YELLOW': {'hex': '#FAFAD2', 'rgb': '(250,250,210)'},
  'LIGHT YELLOW': {'hex': '#FFFFE0', 'rgb': '(255,255,224)'},
  'SADDLE BROWN': {'hex': '#8B4513', 'rgb': '(139,69,19)'},
  'SIENNA': {'hex': '#A0522D', 'rgb': '(160,82,45)'},
  'CHOCOLATE': {'hex': '#D2691E', 'rgb': '(210,105,30)'},
  'PERU': {'hex': '#CD853F', 'rgb': '(205,133,63)'},
  'SANDY BROWN': {'hex': '#F4A460', 'rgb': '(244,164,96)'},
  'BURLY WOOD': {'hex': '#DEB887', 'rgb': '(222,184,135)'},
  'TAN': {'hex': '#D2B48C', 'rgb': '(210,180,140)'},
  'ROSY BROWN': {'hex': '#BC8F8F', 'rgb': '(188,143,143)'},
  'MOCCASIN': {'hex': '#FFE4B5', 'rgb': '(255,228,181)'},
  'NAVAJO WHITE': {'hex': '#FFDEAD', 'rgb': '(255,222,173)'},
  'PEACH PUFF': {'hex': '#FFDAB9', 'rgb': '(255,218,185)'},
  'MISTY ROSE': {'hex': '#FFE4E1', 'rgb': '(255,228,225)'},
  'LAVENDER BLUSH': {'hex': '#FFF0F5', 'rgb': '(255,240,245)'},
  'LINEN': {'hex': '#FAF0E6', 'rgb': '(250,240,230)'},
  'OLD LACE': {'hex': '#FDF5E6', 'rgb': '(253,245,230)'},
  'PAPAYA WHIP': {'hex': '#FFEFD5', 'rgb': '(255,239,213)'},
  'SEA SHELL': {'hex': '#FFF5EE', 'rgb': '(255,245,238)'},
  'MINT CREAM': {'hex': '#F5FFFA', 'rgb': '(245,255,250)'},
  'SLATE GRAY': {'hex': '#708090', 'rgb': '(112,128,144)'},
  'LIGHT SLATE GRAY': {'hex': '#778899', 'rgb': '(119,136,153)'},
  'LIGHT STEEL BLUE': {'hex': '#B0C4DE', 'rgb': '(176,196,222)'},
  'LAVENDER': {'hex': '#E6E6FA', 'rgb': '(230,230,250)'},
  'FLORAL WHITE': {'hex': '#FFFAF0', 'rgb': '(255,250,240)'},
  'ALICE BLUE': {'hex': '#F0F8FF', 'rgb': '(240,248,255)'},
  'GHOST WHITE': {'hex': '#F8F8FF', 'rgb': '(248,248,255)'},
  'HONEYDEW': {'hex': '#F0FFF0', 'rgb': '(240,255,240)'},
  'IVORY': {'hex': '#FFFFF0', 'rgb': '(255,255,240)'},
  'AZURE': {'hex': '#F0FFFF', 'rgb': '(240,255,255)'},
  'SNOW': {'hex': '#FFFAFA', 'rgb': '(255,250,250)'},
  'BLACK': {'hex': '#000000', 'rgb': '(0,0,0)'},
  'DIM GRAY / DIM GREY': {'hex': '#696969', 'rgb': '(105,105,105)'},
  'GRAY / GREY': {'hex': '#808080', 'rgb': '(128,128,128)'},
  'DARK GRAY / DARK GREY': {'hex': '#A9A9A9', 'rgb': '(169,169,169)'},
  'SILVER': {'hex': '#C0C0C0', 'rgb': '(192,192,192)'},
  'LIGHT GRAY / LIGHT GREY': {'hex': '#D3D3D3', 'rgb': '(211,211,211)'},
  'GAINSBORO': {'hex': '#DCDCDC', 'rgb': '(220,220,220)'},
  'WHITE SMOKE': {'hex': '#F5F5F5', 'rgb': '(245,245,245)'},
  'WHITE': {'hex': '#FFFFFF', 'rgb': '(255,255,255)'},
}


def getHexToRgb(hexColor):
  """
  Description:
  ------------
  Convert a hexadecimal color to a rgb code

  An RGB color value is specified with the rgb() function, which has the following syntax:
  rgb(red, green, blue)
  Each parameter (red, green, and blue) defines the intensity of the color and can be an integer between 0 and 255 or a percentage value (from 0% to 100%).
  For example, the rgb(0,0,255) value is rendered as blue, because the blue parameter is set to its highest value (255) and the others are set to 0.
  Also, the following values define equal color: rgb(0,0,255) and rgb(0%,0%,100%).

  Usage::

      ColorMaker().getHexToRgb('#213B68')
  [33, 59, 104]

  https://www.w3schools.com/cssref/css_colors_legal.asp

  Attributes:
  ----------
  :param hexColor: A String with a hexadecimal code color

  :return: The list with the rgb code color
  """
  if not hexColor.startswith("#"):
    raise Exception("Hexadecimal color should start with #")

  if not len(hexColor) == 7:
    raise Exception("Color should have a length of 7")

  return [int(hexColor[1:3], 16), int(hexColor[3:5], 16), int(hexColor[5:7], 16)]


def rgba(red, green, blue, alpha):
  """
  RGBA color values are an extension of RGB color values with an alpha channel - which specifies the opacity of the object.

  An RGBA color is specified with the rgba() function, which has the following syntax:
  rgba(red, green, blue, alpha)
  The alpha parameter is a number between 0.0 (fully transparent) and 1.0 (fully opaque).

  https://www.w3schools.com/cssref/css_colors_legal.asp

  :param red:
  :param green:
  :param blue:
  :param alpha:
  :return:
  """
  return "rgba(%s, %s, %s, %s)" % (red, green, blue, alpha)


def getRgbToHex(rgbColor):
  """
  Description:
  ------------
  Convert a RGB color to a hexadecimal code

  Usage::

      >>> ColorMaker().getRgbToHex([255, 0, 0])
  '#ff0000'

  Attributes:
  ----------
  :param rgbColor: A list corresponding to the RGB color code

  :return: String object defining the hexadecimal color code
  """
  color = []
  for val in rgbColor:
    val = hex(int(val)).lstrip('0x')
    if len(val) != 2:
      leadingZeros = ["0"] * (2 - len(val))
      val = "%s%s" % ("".join(leadingZeros), val)
    color.append(val)
  return "#%s" % "".join(color)


def randColor(seedNo=None):
  """
  Description:
  ------------
  Generate a random hexadecimal color code

  Usage::

      >>> ColorMaker.randColor(10)
  '#9693DD'

  Attributes:
  ----------
  :param seedNo: Optional, The seed number used to generate random numbers

  :return: String with Hexadecimal color code
  """
  letters = '0123456789ABCDEF'
  color = ['#']
  if seedNo is not None:
    random.seed(seedNo)
  for i in range(6):
    color.append(letters[math.floor(random.random() * 16)])
  if seedNo is not None:
    random.seed(None) # Resent the seed to None
  return "".join(color)


def gradient(start, end, factor):
  """
  Description:
  ------------
  Deduce the color from a factor in a range of colors

  Usage::

      >>> ColorMaker().gradient("#ffffff", "#FF0000", 0.2)
  '#ffcccc'

  Attributes:
  ----------
  :param start: The start hexadecimal color code
  :param end: The end hexadecimal color code
  :param factor: A factor in the range [0, 1]

  :return: The hexadecimal color code
  """
  if factor > 1:
    raise Exception("Gradient factor must be <= 1")

  rgbEnd = getHexToRgb(end)
  rgbDiff = [(rgbEnd[i] - val) * factor + val for i, val in enumerate(getHexToRgb(start))]
  return getRgbToHex(rgbDiff)


def colors(start, end, steps):
  """
  Description:
  ------------
  Generate a list of colors between two color codes.

  Usage::

      >>> colors("#ffffff", "#FF0000", 10)
  ['#ffffff', '#ffe2e2', '#ffc6c6', '#ffaaaa', '#ff8d8d', '#ff7171', '#ff5555', '#ff3838', '#ff1c1c', '#FF0000']

  Attributes:
  ----------
  :param start: The start hexadecimal color code
  :param end: The end hexadecimal color code
  :param steps: The number of colors in the array

  :return: A list of hexadecimal color codes
  """
  colors = [start]
  for i in range(steps-2):
    colors.append(gradient(start, end, 1.0 / (steps-1) * (i + 1)))
  colors.append(end)
  return colors


class HexColors(object):
  MAROON = '#800000'
  DARK_RED = '#8B0000'
  BROWN = '#A52A2A'
  FIREBRICK = '#B22222'
  CRIMSON = '#DC143C'
  RED = '#FF0000'
  TOMATO = '#FF6347'
  CORAL = '#FF7F50'
  INDIAN_RED = '#CD5C5C'
  LIGHT_CORAL = '#F08080'
  DARK_SALMON = '#E9967A'
  SALMON = '#FA8072'
  LIGHT_SALMON = '#FFA07A'
  ORANGE_RED = '#FF4500'
  DARK_ORANGE = '#FF8C00'
  ORANGE = '#FFA500'
  GOLD = '#FFD700'
  DARK_GOLDEN_ROD = '#B8860B'
  GOLDEN_ROD = '#DAA520'
  PALE_GOLDEN_ROD = '#EEE8AA'
  DARK_KHAKI = '#BDB76B'
  KHAKI = '#F0E68C'
  OLIVE = '#808000'
  YELLOW = '#FFFF00'
  YELLOW_GREEN = '#9ACD32'
  DARK_OLIVE_GREEN = '#556B2F'
  OLIVE_DRAB = '#6B8E23'
  LAWN_GREEN = '#7CFC00'
  CHART_REUSE = '#7FFF00'
  GREEN_YELLOW = '#ADFF2F'
  DARK_GREEN = '#006400'
  GREEN = '#008000'
  FOREST_GREEN = '#228B22'
  LIME = '#00FF00'
  LIME_GREEN = '#32CD32'
  LIGHT_GREEN = '#90EE90'
  PALE_GREEN = '#98FB98'
  DARK_SEA_GREEN = '#8FBC8F'
  MEDIUM_SPRING_GREEN = '#00FA9A'
  SPRING_GREEN = '#00FF7F'
  SEA_GREEN = '#2E8B57'
  MEDIUM_AQUA_MARINE = '#66CDAA'
  MEDIUM_SEA_GREEN = '#3CB371'
  LIGHT_SEA_GREEN = '#20B2AA'
  DARK_SLATE_GRAY = '#2F4F4F'
  TEAL = '#008080'
  DARK_CYAN = '#008B8B'
  AQUA = '#00FFFF'
  CYAN = '#00FFFF'
  LIGHT_CYAN = '#E0FFFF'
  DARK_TURQUOISE = '#00CED1'
  TURQUOISE = '#40E0D0'
  MEDIUM_TURQUOISE = '#48D1CC'
  PALE_TURQUOISE = '#AFEEEE'
  AQUA_MARINE = '#7FFFD4'
  POWDER_BLUE = '#B0E0E6'
  CADET_BLUE = '#5F9EA0'
  STEEL_BLUE = '#4682B4'
  CORN_FLOWER_BLUE = '#6495ED'
  DEEP_SKY_BLUE = '#00BFFF'
  DODGER_BLUE = '#1E90FF'
  LIGHT_BLUE = '#ADD8E6'
  SKY_BLUE = '#87CEEB'
  LIGHT_SKY_BLUE = '#87CEFA'
  MIDNIGHT_BLUE = '#191970'
  NAVY = '#000080'
  DARK_BLUE = '#00008B'
  MEDIUM_BLUE = '#0000CD'
  BLUE = '#0000FF'
  ROYAL_BLUE = '#4169E1'
  BLUE_VIOLET = '#8A2BE2'
  INDIGO = '#4B0082'
  DARK_SLATE_BLUE = '#483D8B'
  SLATE_BLUE = '#6A5ACD'
  MEDIUM_SLATE_BLUE = '#7B68EE'
  MEDIUM_PURPLE = '#9370DB'
  DARK_MAGENTA = '#8B008B'
  DARK_VIOLET = '#9400D3'
  DARK_ORCHID = '#9932CC'
  MEDIUM_ORCHID = '#BA55D3'
  PURPLE = '#800080'
  THISTLE = '#D8BFD8'
  PLUM = '#DDA0DD'
  VIOLET = '#EE82EE'
  ORCHID = '#DA70D6'
  MEDIUM_VIOLET_RED = '#C71585'
  PALE_VIOLET_RED = '#DB7093'
  DEEP_PINK = '#FF1493'
  HOT_PINK = '#FF69B4'
  LIGHT_PINK = '#FFB6C1'
  PINK = '#FFC0CB'
  ANTIQUE_WHITE = '#FAEBD7'
  BEIGE = '#F5F5DC'
  BISQUE = '#FFE4C4'
  BLANCHED_ALMOND = '#FFEBCD'
  WHEAT = '#F5DEB3'
  CORN_SILK = '#FFF8DC'
  LEMON_CHIFFON = '#FFFACD'
  LIGHT_GOLDEN_ROD_YELLOW = '#FAFAD2'
  LIGHT_YELLOW = '#FFFFE0'
  SADDLE_BROWN = '#8B4513'
  SIENNA = '#A0522D'
  CHOCOLATE = '#D2691E'
  PERU = '#CD853F'
  SANDY_BROWN = '#F4A460'
  BURLY_WOOD = '#DEB887'
  TAN = '#D2B48C'
  ROSY_BROWN = '#BC8F8F'
  MOCCASIN = '#FFE4B5'
  NAVAJO_WHITE = '#FFDEAD'
  PEACH_PUFF = '#FFDAB9'
  MISTY_ROSE = '#FFE4E1'
  LAVENDER_BLUSH = '#FFF0F5'
  LINEN = '#FAF0E6'
  OLD_LACE = '#FDF5E6'
  PAPAYA_WHIP = '#FFEFD5'
  SEA_SHELL = '#FFF5EE'
  MINT_CREAM = '#F5FFFA'
  SLATE_GRAY = '#708090'
  LIGHT_SLATE_GRAY = '#778899'
  LIGHT_STEEL_BLUE = '#B0C4DE'
  LAVENDER = '#E6E6FA'
  FLORAL_WHITE = '#FFFAF0'
  ALICE_BLUE = '#F0F8FF'
  GHOST_WHITE = '#F8F8FF'
  HONEYDEW = '#F0FFF0'
  IVORY = '#FFFFF0'
  AZURE = '#F0FFFF'
  SNOW = '#FFFAFA'
  BLACK = '#000000'
  SILVER = '#C0C0C0'
  GAINSBORO = '#DCDCDC'
  WHITE_SMOKE = '#F5F5F5'
  WHITE = '#FFFFFF'


class RgbColors(object):
  TRANSPARENT = "rgba(0, 0, 0, 0)"
  MAROON = 'rgb(128,0,0)'
  DARK_RED = 'rgb(139,0,0)'
  BROWN = 'rgb(165,42,42)'
  FIREBRICK = 'rgb(178,34,34)'
  CRIMSON = 'rgb(220,20,60)'
  RED = 'rgb(255,0,0)'
  TOMATO = 'rgb(255,99,71)'
  CORAL = 'rgb(255,127,80)'
  INDIAN_RED = 'rgb(205,92,92)'
  LIGHT_CORAL = 'rgb(240,128,128)'
  DARK_SALMON = 'rgb(233,150,122)'
  SALMON = 'rgb(250,128,114)'
  LIGHT_SALMON = 'rgb(255,160,122)'
  ORANGE_RED = 'rgb(255,69,0)'
  DARK_ORANGE = 'rgb(255,140,0)'
  ORANGE = 'rgb(255,165,0)'
  GOLD = 'rgb(255,215,0)'
  DARK_GOLDEN_ROD = 'rgb(184,134,11)'
  GOLDEN_ROD = 'rgb(218,165,32)'
  PALE_GOLDEN_ROD = 'rgb(238,232,170)'
  DARK_KHAKI = 'rgb(189,183,107)'
  KHAKI = 'rgb(240,230,140)'
  OLIVE = 'rgb(128,128,0)'
  YELLOW = 'rgb(255,255,0)'
  YELLOW_GREEN = 'rgb(154,205,50)'
  DARK_OLIVE_GREEN = 'rgb(85,107,47)'
  OLIVE_DRAB = 'rgb(107,142,35)'
  LAWN_GREEN = 'rgb(124,252,0)'
  CHART_REUSE = 'rgb(127,255,0)'
  GREEN_YELLOW = 'rgb(173,255,47)'
  DARK_GREEN = 'rgb(0,100,0)'
  GREEN = 'rgb(0,128,0)'
  FOREST_GREEN = 'rgb(34,139,34)'
  LIME = 'rgb(0,255,0)'
  LIME_GREEN = 'rgb(50,205,50)'
  LIGHT_GREEN = 'rgb(144,238,144)'
  PALE_GREEN = 'rgb(152,251,152)'
  DARK_SEA_GREEN = 'rgb(143,188,143)'
  MEDIUM_SPRING_GREEN = 'rgb(0,250,154)'
  SPRING_GREEN = 'rgb(0,255,127)'
  SEA_GREEN = 'rgb(46,139,87)'
  MEDIUM_AQUA_MARINE = 'rgb(102,205,170)'
  MEDIUM_SEA_GREEN = 'rgb(60,179,113)'
  LIGHT_SEA_GREEN = 'rgb(32,178,170)'
  DARK_SLATE_GRAY = 'rgb(47,79,79)'
  TEAL = 'rgb(0,128,128)'
  DARK_CYAN = 'rgb(0,139,139)'
  AQUA = 'rgb(0,255,255)'
  CYAN = 'rgb(0,255,255)'
  LIGHT_CYAN = 'rgb(224,255,255)'
  DARK_TURQUOISE = 'rgb(0,206,209)'
  TURQUOISE = 'rgb(64,224,208)'
  MEDIUM_TURQUOISE = 'rgb(72,209,204)'
  PALE_TURQUOISE = 'rgb(175,238,238)'
  AQUA_MARINE = 'rgb(127,255,212)'
  POWDER_BLUE = 'rgb(176,224,230)'
  CADET_BLUE = 'rgb(95,158,160)'
  STEEL_BLUE = 'rgb(70,130,180)'
  CORN_FLOWER_BLUE = 'rgb(100,149,237)'
  DEEP_SKY_BLUE = 'rgb(0,191,255)'
  DODGER_BLUE = 'rgb(30,144,255)'
  LIGHT_BLUE = 'rgb(173,216,230)'
  SKY_BLUE = 'rgb(135,206,235)'
  LIGHT_SKY_BLUE = 'rgb(135,206,250)'
  MIDNIGHT_BLUE = 'rgb(25,25,112)'
  NAVY = 'rgb(0,0,128)'
  DARK_BLUE = 'rgb(0,0,139)'
  MEDIUM_BLUE = 'rgb(0,0,205)'
  BLUE = 'rgb(0,0,255)'
  ROYAL_BLUE = 'rgb(65,105,225)'
  BLUE_VIOLET = 'rgb(138,43,226)'
  INDIGO = 'rgb(75,0,130)'
  DARK_SLATE_BLUE = 'rgb(72,61,139)'
  SLATE_BLUE = 'rgb(106,90,205)'
  MEDIUM_SLATE_BLUE = 'rgb(123,104,238)'
  MEDIUM_PURPLE = 'rgb(147,112,219)'
  DARK_MAGENTA = 'rgb(139,0,139)'
  DARK_VIOLET = 'rgb(148,0,211)'
  DARK_ORCHID = 'rgb(153,50,204)'
  MEDIUM_ORCHID = 'rgb(186,85,211)'
  PURPLE = 'rgb(128,0,128)'
  THISTLE = 'rgb(216,191,216)'
  PLUM = 'rgb(221,160,221)'
  VIOLET = 'rgb(238,130,238)'
  ORCHID = 'rgb(218,112,214)'
  MEDIUM_VIOLET_RED = 'rgb(199,21,133)'
  PALE_VIOLET_RED = 'rgb(219,112,147)'
  DEEP_PINK = 'rgb(255,20,147)'
  HOT_PINK = 'rgb(255,105,180)'
  LIGHT_PINK = 'rgb(255,182,193)'
  PINK = 'rgb(255,192,203)'
  ANTIQUE_WHITE = 'rgb(250,235,215)'
  BEIGE = 'rgb(245,245,220)'
  BISQUE = 'rgb(255,228,196)'
  BLANCHED_ALMOND = 'rgb(255,235,205)'
  WHEAT = 'rgb(245,222,179)'
  CORN_SILK = 'rgb(255,248,220)'
  LEMON_CHIFFON = 'rgb(255,250,205)'
  LIGHT_GOLDEN_ROD_YELLOW = 'rgb(250,250,210)'
  LIGHT_YELLOW = 'rgb(255,255,224)'
  SADDLE_BROWN = 'rgb(139,69,19)'
  SIENNA = 'rgb(160,82,45)'
  CHOCOLATE = 'rgb(210,105,30)'
  PERU = 'rgb(205,133,63)'
  SANDY_BROWN = 'rgb(244,164,96)'
  BURLY_WOOD = 'rgb(222,184,135)'
  TAN = 'rgb(210,180,140)'
  ROSY_BROWN = 'rgb(188,143,143)'
  MOCCASIN = 'rgb(255,228,181)'
  NAVAJO_WHITE = 'rgb(255,222,173)'
  PEACH_PUFF = 'rgb(255,218,185)'
  MISTY_ROSE = 'rgb(255,228,225)'
  LAVENDER_BLUSH = 'rgb(255,240,245)'
  LINEN = 'rgb(250,240,230)'
  OLD_LACE = 'rgb(253,245,230)'
  PAPAYA_WHIP = 'rgb(255,239,213)'
  SEA_SHELL = 'rgb(255,245,238)'
  MINT_CREAM = 'rgb(245,255,250)'
  SLATE_GRAY = 'rgb(112,128,144)'
  LIGHT_SLATE_GRAY = 'rgb(119,136,153)'
  LIGHT_STEEL_BLUE = 'rgb(176,196,222)'
  LAVENDER = 'rgb(230,230,250)'
  FLORAL_WHITE = 'rgb(255,250,240)'
  ALICE_BLUE = 'rgb(240,248,255)'
  GHOST_WHITE = 'rgb(248,248,255)'
  HONEYDEW = 'rgb(240,255,240)'
  IVORY = 'rgb(255,255,240)'
  AZURE = 'rgb(240,255,255)'
  SNOW = 'rgb(255,250,250)'
  BLACK = 'rgb(0,0,0)'
  SILVER = 'rgb(192,192,192)'
  GAINSBORO = 'rgb(220,220,220)'
  WHITE_SMOKE = 'rgb(245,245,245)'
  WHITE = 'rgb(255,255,255)'


class DefinedColors(object):

  @property
  def hex(self):
    """
    Description:
    ------------
    Returns the Hexadecimal predefined color codes

    Related Pages:

			https://www.rapidtables.com/web/color/RGB_Color.html
    """
    return HexColors

  @property
  def rgb(self):
    """
    Description:
    ------------
    Returns the RGB predefined color codes

    Related Pages:

			https://www.rapidtables.com/web/color/RGB_Color.html
    """
    return RgbColors




from epyk.core.Page import Report
from epyk.core.py import Pyk
rptObj = Report()


button = rptObj.ui.buttons.button('This is button')
imp = rptObj.ui.buttons.important('This is an important button')

Pyk.exports({'button': button, 'imp': imp})



from epyk.core.html import Html
from epyk.core.html.options import OptVis

from epyk.core.js import JsUtils
from epyk.core.css.styles import GrpClsCharts
from epyk.core.js.packages import JsVis


class Chart(Html.Html):
  name = 'Vis'
  requirements = ('vis', )

  def __init__(self, report, width, height, htmlCode, options, profile):
    self.height = height[0]
    super(Chart, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._d3, self._datasets, self._options, self._data_attrs, self._attrs = None, [], None, {}, {}
    self._options_init = options
    self.style.css.margin_top = 10
    self.style.css.display = "inline-block"

  @property
  def chartId(self):
    """
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def options(self):
    """

    :rtype: OptVis.Options2D
    """
    if self._options is None:
      self._options = OptVis.Options2D(self._report, attrs=self._options_init)
    return self._options

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object

    :rtype: JsVis.VisGraph2D
    """
    if self._js is None:
      self._js = JsVis.VisGraph2D(selector="window['%s']" % self.chartId, src=self)
    return self._js

  def build(self, data=None, options=None, profile=False):
    if data:
      return JsUtils.jsConvertFncs([self.js.setItems(data[0]), self.js.redraw()], toStr=True)

    return '''%s; var %s = %s''' % (self.groups.toStr(), self.chartId, self.getCtx())

  def getCtx(self):
    raise Exception("Cannot create an object from the Vis base class directly")

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())


class ChartLine(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartLine, self).__init__(report, width, height, htmlCode, options, profile)
    self.items, self.__grps = [], None
    self.options.style = "line"

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object

    :rtype: JsVis.VisGraph2D
    """
    if self._js is None:
      self._js = JsVis.VisGraph2D(selector="window['%s']" % self.chartId, src=self)
    return self._js

  @property
  def groups(self):
    """

    :rtype: JsVis.VisGroups
    """
    if self.__grps is None:
      self.__grps = JsVis.VisGroups(self._report, setVar=True, varName="%s_group" % self.chartId)
    return self.__grps

  def add_item(self, x, y, group, label=None, end=None):
    """

    :param x:
    :param y:
    :param group:
    :param label:
    :param end:
    """
    self.items.append({"x": x, "y": y, "group": group})
    return self

  def add_items(self, records):
    for rec in records:
      self.add_item(rec['x'], rec['y'], rec.get('group', 0))
    return self

  def getCtx(self):
    return "new vis.Graph2d(%s, %s, %s, %s)" % (self.dom.varId, self.items, self.groups.varId, self.options)


class ChartBar(ChartLine):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartBar, self).__init__(report, width, height, htmlCode, options, profile)
    self.options.style = "bar"


class ChartScatter(ChartLine):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartScatter, self).__init__(report, width, height, htmlCode, options, profile)
    self.options.style = "points"
    self.options.drawPoints.style.circle()


class Chart3D(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(Chart3D, self).__init__(report, width, height, htmlCode, options, profile)
    self.items, self.__grps = [], None
    self.options.style.surface()
    self.options.showPerspective = True
    self.options.showGrid = True
    self.options.keepAspectRatio = True
    self.options.verticalRatio = 0.5

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object

    :rtype: JsVis.VisGraph2D
    """
    if self._js is None:
      self._js = JsVis.VisGraph3D(selector="window['%s']" % self.chartId, src=self)
    return self._js

  @property
  def options(self):
    """

    :rtype: OptVis.Options3D
    """
    if self._options is None:
      self._options = OptVis.Options3D(self._report, attrs=self._options_init)
    return self._options

  @property
  def groups(self):
    """

    :rtype: JsVis.VisGroups
    """
    if self.__grps is None:
      self.__grps = JsVis.VisGroups(self._report, setVar=True, varName="%s_group" % self.chartId)
    return self.__grps

  @property
  def js(self):
    """

    :rtype: JsVis.VisGraph3D
    """

    if self._js is None:
      self._js = JsVis.VisGraph3D(self._report, varName=self.chartId)
    return self._js

  def add_items(self, records):
    for rec in records:
      self.add_item(rec['x'], rec['y'], rec['z'], rec.get('group', 0))
    return self

  def add_item(self, x, y, z, group=0, style=None):
    """

    :param x:
    :param y:
    :param z:
    :param style:
    """
    self.items.append({"x": x, "y": y, "z": z, 'group': group})
    return self

  def build(self, data=None, options=None, profile=False):
    if data:
      return "%(chartId)s.setData(%(data)s); %(chartId)s.redraw()" % {'chartId': self.chartId, 'data': data[0]}

    return '''%s; var %s = %s''' % (self.groups.toStr(), self.chartId, self.getCtx())

  def getCtx(self):
    return "new vis.Graph3d(%s, %s, %s)" % (self.dom.varId, self.items, self.options)


class Chart3DScatter(Chart3D):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(Chart3DScatter, self).__init__(report, width, height, htmlCode, options, profile)
    self.items, self.__grps = [], None
    self.options.style.dot()


class Chart3DLine(Chart3D):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(Chart3DLine, self).__init__(report, width, height, htmlCode, options, profile)
    self.items, self.__grps = [], None
    self.options.style.line()


class Chart3DBar(Chart3D):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(Chart3DBar, self).__init__(report, width, height, htmlCode, options, profile)
    self.items, self.__grps = [], None
    self.options.style.bar()


class ChartNetwork(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartNetwork, self).__init__(report, width, height, htmlCode, options, profile)
    self._nodes, self._edges, self.__grps = [], [], None

  @property
  def options(self):
    """

    :rtype: Options2D
    """
    if self._options is None:
      self._options = OptVis.OptionsNetwork(self._report, attrs=self._options_init)
    return self._options

  @property
  def js(self):
    """
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object
    :rtype: Js.JsBase
    """
    if self._js is None:
      self._js = JsVis.VisNetwork(self._report, selector=self.chartId, setVar=False)
    return self._js

  @property
  def groups(self):
    """

    :return:
    """
    if self.__grps is None:
      self.__grps = JsVis.VisGroups(self._report, setVar=True, varName="%s_group" % self.chartId)
    return self.__grps

  def add_node(self, label, id=None, group=0):
    """

    :param label:
    :param id:
    :param group:
    """
    self._nodes.append({'label': label, 'group': group, 'id': id or len(self._nodes)})
    return self

  def add_edge(self, fromId, toId, title=None, relation=None):
    self._edges.append({'from': fromId, 'to': toId})
    return self

  def getCtx(self):
    return "new vis.Network(%s, %s, %s, %s)" % (self.dom.varId, {"nodes": self._nodes, 'edges': self._edges}, self.groups.varId, self.options)


class ChartTimeline(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartTimeline, self).__init__(report, width, height, htmlCode, options, profile)
    self.items, self.__grps, self.__cats = [], None, None

  @property
  def style(self):
    if self._styleObj is None:
      self._styleObj = GrpClsCharts.ClassVisTimeline(self)
    return self._styleObj

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: Options2D
    """
    if self._options is None:
      self._options = OptVis.OptionsTimeline(self._report, attrs=self._options_init)
    return self._options

  @property
  def js(self):
    """
    Description:
    ------------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object

    :rtype: Js.JsBase
    """
    if self._js is None:
      self._js = JsVis.VisTimeline(selector=self.chartId, src=self)
    return self._js

  @property
  def groups(self):
    """
    Description:
    ------------

    """
    if self.__grps is None:
      self.__grps = JsVis.VisGroups(self._report, setVar=True, varName="%s_group" % self.chartId)
    return self.__grps

  def setGroups(self, groups):
    """
    Description:
    ------------
    Add group labels to the timeline.

    Attributes:
    ----------
    :param groups:
    """
    self.__cats = "%s.setGroups(%s)" % (self.chartId, JsUtils.jsConvertData(groups, None))
    return self

  def add_items(self, records):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param records:
    """
    self.items.extend(records)
    return self

  def getCtx(self):
    if self.__cats is not None:
      return "new vis.Timeline(%s, %s, %s); %s" % (self.dom.varId, self.items, self.options, self.__cats)

    return "new vis.Timeline(%s, %s, %s)" % (self.dom.varId, self.items, self.options)



"""
Wrapper to the Javascript require package
This module will replace the Python Import module if specified

Related Pages:

			https://requirejs.org/docs/api.html#mechanics
"""

from epyk.core.js.packages import JsPackage


class JsRequire(JsPackage):
  def resolveImports(self, modules, moduleStack, level, modulesBase):
    """

    :param modules:
    :param moduleStack:
    :param level:

    :return:
    """
    for m in modules:
      mDef = UIDeps.PACKAGES.get(m, {})
      reqModules = []
      for req in mDef.get('req', []):
        reqModules.append(req['alias'])
      if len(reqModules) > 0:
        moduleStack.append([])
        self.resolveImports(reqModules, moduleStack, level+1, modulesBase)
      else:
        modulesBase.append(m)
      moduleStack[level].append(m)

  def getImports(self, modules):
    result = {'header': [], 'requireJs': ''}
    moduleStack, modulesBase = [[]], []
    self.resolveImports(modules, moduleStack, 0, modulesBase)
    moduleStack[-1].extend(modulesBase)
    mOrdered, reqJs = [], []
    for mLevel in moduleStack[::-1]:
      reqJs.append([])
      for m in mLevel:
        if m not in mOrdered:
          mOrdered.append(m)
          reqJs[-1].append(m)
      if len(reqJs[-1]) == 0:
        reqJs.pop()
    return {'header': mOrdered, 'reqJs': reqJs}

  def fullHeader(self, modules, repoPath=None):
    header = []
    for alias in modules:
      mDef = UIDeps.PACKAGES.get(alias, {})
      for m in mDef.get('modules', []):
        path = m["path"] % m
        if repoPath is None:
          fullPath = '%s/%s%s' % (mDef['cdnjs'], path, m['script'])
        else:
          fullPath = '%s/%s%s' % (repoPath, path, m['script'])
        if m['script'].endswith(".css"):
          header.append("<link rel='stylesheet' href='%s'>" % fullPath)
        else:
          header.append("<script  src='%s' type='text/javascript'=></script>" % fullPath)
    return header



from epyk.core.js.html import JsHtml
from epyk.core.js import JsUtils

class HtmlComp(object):
  htmlCode = "test"


# ------------------------------------------------------------------------------------------------------------------
#
dom = JsHtml.JsHtml(HtmlComp())
f = JsUtils.JsFile("Dom", path=r"../outs")

f.writeJs([
  dom.hide(),
  dom.toggle()
])
f.close()




import re

from epyk.core.html import Html
from epyk.core.html import Defaults
from epyk.core.html.options import OptButton

from epyk.core.css import Defaults_css
from epyk.core.css.styles import GrpClsImage

# The list of Javascript classes
from epyk.core.js.html import JsHtml


class Image(Html.Html):
  name = 'Picture'

  def __init__(self, report, image, path, align, htmlCode, width, height, profile, options):
    if path is None:
      path = Defaults.SERVER_PATH if report.run.report_name is None else "%s/%s" % (Defaults.SERVER_PATH, report.run.report_name)
    super(Image, self).__init__(report, {'path': path, 'image': image}, htmlCode=htmlCode, profile=profile,
                                css_attrs={"width": width, "height": height})
    self._jsStyles = options
    if align is not None:
      self.css({"text-align": align})

  @property
  def _js__builder__(self):
    return '''
      if(typeof data.path === 'undefined'){data.path = '%s'};
      htmlObj.src = data.path + "/" + data.image; 
      if(typeof options.css !== 'undefined'){for(var k in options.css){htmlObj.style[k] = options.css[k]}}
      ''' % Defaults.SERVER_PATH

  def __str__(self):
    self.attr["src"] = "%(path)s/%(image)s" % self.val
    return '<img %s></img>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class AnimatedImage(Html.Html):
  name = 'Animated Picture'

  def __init__(self, report, image, text, title, url, path, width, height, profile):
    if path is None:
      path = Defaults.SERVER_PATH if report.run.report_name is None else "%s/%s" % (Defaults.SERVER_PATH, report.run.report_name)
    super(AnimatedImage, self).__init__(report, {'path': path, 'image': image, 'text': text, "title": title, 'url': url},
                                        css_attrs={"width": width, "height": height, 'overflow': 'hidden', 'display': 'block'}, profile=profile)
    self.img = report.ui.img(image, path=path, width=(width[0]-5, width[1]), height=("auto", ''))
    self.img.options.managed = False
    self.title = report.ui.tags.h2(title).css({"display": 'block'})
    self.text = report.ui.tags.p(text).css({"display": 'block'})
    self.a = report.ui.tags.a("Enter", url).css({"width": "100px"})
    self.a.style.add_classes.image.info_link()
    self.div = report.ui.div([self.title, self.text, self.a], width=(width[0]-2, width[1])).css({"padding": "5px"})
    self.div.style.add_classes.image.mask()
    self.div.options.managed = False

  def __str__(self):
    return '''<div %(cssAttr)s>%(div)s%(img)s</div>
      ''' % {"cssAttr": self.get_attrs(pyClassNames=self.style.get_classes()), 'img': self.img.html(), 'div': self.div.html()}


class ImgCarrousel(Html.Html):
  name = 'Carrousel'

  def __init__(self, report, images, path, selected, width, height, profile):
    if path is None:
      path = Defaults.SERVER_PATH if report.run.report_name is None else "%s/%s" % (Defaults.SERVER_PATH, report.run.report_name)
    self.items, self.__click_items = [], []
    super(ImgCarrousel, self).__init__(report, "", css_attrs={"width": width, "height": height}, profile=profile)
    for i, rec in enumerate(images):
      if not isinstance(rec, dict):
        rec = {"image": rec, 'title': "picture %s" % (i+1)}
      if not 'path' in rec:
        rec['path'] = path
      if rec.get('selected') is not None:
        selected = i
      img = report.ui.img(rec["image"], path=rec["path"], width=width, height=(height[0] - 60, height[1]))
      div = report.ui.layouts.div([report.ui.tags.h3(rec['title']), img], htmlCode="%s_img_%s" % (self.htmlCode, i)).css({"display": 'none', "text-align": "center"})
      div.set_attrs(name="name", value="%s_img" % self.htmlCode)
      div.options.managed = False
      self.items.append(div)
    self.items[selected].css({"display": 'block'})
    self.css({'padding-top': '20px', 'padding': "2px", 'margin': 0})

  def __getitem__(self, i):
    return self.items[i]

  def click(self, jsFncs, profile=False, source_event=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__click_items.extend(jsFncs)
    return self

  @property
  def _js__builder__(self):
    return '''
      data.forEach(function(rec, i){
        var li = document.createElement('li');
        if (i == 0) {li.style.display = 'block'} else{li.style.display = 'none'};
        var img = document.createElement('img'); img.src = rec.path +'/'+ rec.image; li.appendChild(img);
        var title = document.createElement('h3'); title.innerHTML = rec.title; li.appendChild(title); 
        htmlObj.appendChild(li);
        var label = document.createElement('label'); label.style.backgroundColor = '%(color)s'; 
        label.style.borderRadius = '20px'; label.for = i; label.innerHTML = '&nbsp;'; 
        document.getElementById(htmlObj.id +'_bullets').appendChild(label)
      })''' % {'color': self._report.theme.colors[9]}

  def __str__(self):
    img_cont = self._report.ui.layouts.div(self.items).css({"display": 'block', "width": "100%", "text-align": "center"})
    img_cont.options.managed = False
    points = self._report.ui.navigation.points(len(self.items))
    points.options.managed = False
    points.click([
      self._report.js.getElementsByName("%s_img" % self.htmlCode).css({"display": 'none'}),
      self._report.js.getElementById("%s_img_' + data.position +'" % self.htmlCode).css({"display": 'block'})
    ] + self.__click_items)
    return '''<div %(strAttr)s>%(img_cont)s%(points)s</div>
      ''' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()),
             'img_cont': img_cont.html(), "points": points.html()}


class Icon(Html.Html):
  requirements = ('font-awesome', )
  name = 'Icon'

  def __init__(self, report, value, width, height, color, tooltip, profile):
    super(Icon, self).__init__(report, "", css_attrs={"color": color, "width": width, "height": height}, profile=profile)
    if value is not None:
      self.attr['class'].add(value)
    self.attr['aria-hidden'] = 'true'
    if tooltip is not None:
      self.set_attrs(name="title", value=tooltip)

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlIcon
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlIcon(self, report=self._report)
    return self._dom

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsImage.ClassIcon
    """
    if self._styleObj is None:
      self._styleObj = GrpClsImage.ClassIcon(self)
    return self._styleObj

  def set_icon(self, value):
    """

    :param value:

    :return:
    """
    self.attr['class'].add(value)
    return self

  def hover_colors(self, color_hover, color_out=None):
    """
    Description:
    ------------
    Change the color of the button background when the mouse is hover

    Usage::

      rptObj.ui.icons.capture().icon.hover_colors("red", "yellow")

    Attributes:
    ----------
    :param color_hover: String. The color of the icon when mouse hover
    :param color_out: Optional, String. The color of the icon when mouse out
    """
    if color_out is None:
      color_out = self._report.theme.success[1]
    else:
      self.css({"color": color_out})
    self.set_attrs(name="onmouseover", value="this.style.color='%s'" % color_hover)
    self.set_attrs(name="onmouseout", value="this.style.color='%s'" % color_out)
    return self

  def click(self, jsFncs, profile=False, source_event=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    :return:
    """
    self.style.css.cursor = "pointer"
    return super(Icon, self).click(jsFncs, profile, source_event)

  @property
  def _js__builder__(self):
    return '''htmlObj.classList = []; data.split(' ').forEach(function(cls){htmlObj.classList.add(cls)});
      if(typeof options.css !== 'undefined'){for(var k in options.css){htmlObj.style[k] = options.css[k]}}'''

  def __str__(self):
    return '<i %s></i>' % (self.get_attrs(pyClassNames=self.style.get_classes()))


class Emoji(Html.Html):
  name = 'Emoji'

  def __init__(self, report, symbole, top, profile):
    super(Emoji, self).__init__(report, symbole, profile=profile)
    self.style.css.margin_top = '%s%s' % (top[0], top[1])

  @property
  def _js__builder__(self):
    return '''
      htmlObj.innerHTML = data; 
      if(typeof options.css !== 'undefined'){for(var k in options.css){htmlObj.style[k] = options.css[k]}}'''

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlRich
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlRich(self, report=self._report)
    return self._dom

  def __str__(self):
    return '<p %s>%s</p>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val)


class Badge(Html.Html):
  name = 'Badge'
  requirements = ('font-awesome', 'bootstrap')

  def __init__(self, report, text, width, height, label, icon, background_color, color, url, tooltip, options, profile):
    super(Badge, self).__init__(report, None, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_label(label, css={"vertical-align": "middle", "width": 'none', "height": 'none'})
    self.__options = OptButton.OptionsBadge(self, options)
    if self.options.badge_position == 'left':
      self.add_icon(icon, css={"float": 'None', 'margin-left': "5px"}, position="after")
    else:
      self.add_icon(icon, css={"float": 'left', 'margin-left': "5px"})
    if hasattr(self.icon, 'css') and width[0] is not None:
      self.icon.css({"font-size": "%s%s" % (width[0], width[1])})
    self.link = None
    if url is not None:
      self.link = self._report.ui.links.external(text, url).css({"color": "inherit", 'display': 'inline-block',
          "padding": "2px", "width": "auto"})
      self.link.options.managed = False
    else:
      self.link = self._report.ui.text(text).css({'display': 'inline-block',
          "padding": "2px", "width": "auto"})
    self.link.css(self.options.badge_css)
    self.link.css({"color": color, "border-radius": "20px", 'margin-left': '2px', 'position': 'relative', 'right': '12px',
                   'background': background_color, 'top': "-5px"})
    self.link.options.managed = False
    self.attr['class'].add("badge") # From bootstrap
    if tooltip is not None:
      self.tooltip(tooltip)

  @property
  def options(self):
    """
    Description:
    ------------
    Property to the options specific to the HTML component

    :rtype: OptButton.OptionsBadge
    """
    return self.__options

  def __str__(self):
    return '<span %s>%s</span>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.link)




from epyk.core import html
from epyk.core.css import Defaults as Defaults_css


class Vignets(object):
  def __init__(self, context):
    self.context = context

  def bubble(self, recordSet=None, width=(50, "px"), height=(110, 'px'), color=None,
             background_color=None, helper=None, profile=None):
    """
    Description:
    ------------
    The bubbles event property returns a Boolean value that indicates whether or not an event is a bubbling event.
    Event bubbling directs an event to its intended target, it works like this:
    A button is clicked and the event is directed to the button
    If an event handler is set for that object, the event is triggered
    If no event handler is set for that object, the event bubbles up (like a bubble in water) to the objects parent
    The event bubbles up from parent to parent until it is handled, or until it reaches the document object.

    Usage::

      rptObj.ui.vignets.bubble({"value": 23, "title": "Title"}, helper="This is a helper")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Text`
      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlLinks.ExternalLink`

    Related Pages:

			https://www.w3schools.com/jsref/event_bubbles.asp

    Attributes:
    ----------
    :param recordSet:
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param color: Optional. The font color in the component. Default inherit
    :param background_color:
    :param helper: Optional. A tooltip helper
    :param profile: Optional. A flag to set the component performance storage
    """
    div = self.context.rptObj.ui.div(width=width, height=height)
    bubble = self.context.rptObj.ui.div(width=width, height=(height[0]-60, height[1]))
    div.number = self.context.rptObj.ui.text(recordSet["value"], width=width)
    if recordSet.get("url") is not None:
      div.title = self.context.rptObj.ui.link(recordSet["title"], url=recordSet['url'])
      div.title.no_decoration()
    else:
      div.title = self.context.rptObj.ui.text(recordSet["title"])
    div.title.style.css.bold()
    div.number.style.css.line_height = height[0]-60
    div.number.style.css.text_align = "center"
    div.number.style.css.font_size = height[0]-90
    bubble += div.number
    bubble.style.css.background_color = self.context.rptObj.theme.success[1]
    bubble.style.css.color = self.context.rptObj.theme.greys[0]
    bubble.style.css.borders_light()
    bubble.style.css.border_radius = height[0]-60
    bubble.style.css.middle()
    div.style.css.text_align = "center"
    div += bubble
    div += div.title
    return div

  def number(self, number, label="", width=(100, "px"), height=(None, "px"), profile=None, options=None):
    """
    Description:
    ------------
    The <input type="number"> defines a field for entering a number.
    Use the following attributes to specify restrictions:
    max - specifies the maximum value allowed
    min - specifies the minimum value allowed
    step - specifies the legal number intervals
    value - Specifies the default value

    Usage::

      number = rptObj.ui.vignets.number(500, "Test")
      number_2 = rptObj.ui.vignets.number(500, "Test 2 ", options={"url": "http://www.google.fr"})
      number.span.add_icon(rptObj.ui.icons.get.ICON_ENVELOPE)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextComp.Number`

    Related Pages:

			https://www.w3schools.com/tags/att_input_type_number.asp

    Attributes:
    ----------
    :param number:
    :param label:
    :param width:
    :param height:
    :param profile:
    """
    html_number = html.HtmlTextComp.Number(self.context.rptObj, number, label, width, height, profile, options or {})
    return html_number

  def link(self):
    pass

  def block(self, recordSet=None, color=None, border='auto', width=(300, 'px'), height=(None, 'px'),
            helper=None, options=None, profile=None):
    """
    Description:
    ------------
    Every HTML element has a default display value depending on what type of element it is.
    The two display values are: block and inline.

    Usage::

      rptObj.ui.vignets.block({"text": 'This is a brand new python framework', "title": 'New Python Web Framework',
                             "button": {"text": 'Get Started', 'url': "/getStarted"}, 'color': 'green'})

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextComp.BlockText`

    Related Pages:

			https://www.w3schools.com/htmL/html_blocks.asp

    Attributes:
    ----------
    :param recordSet:
    :param color: Optional. The font color in the component. Default inherit
    :param border:
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param helper: Optional. A tooltip helper
    :param profile: Optional. A flag to set the component performance storage
    """
    html_blocktext = html.HtmlTextComp.BlockText(self.context.rptObj, recordSet, color, border, width, height, helper, options, profile)
    return html_blocktext

  def text(self, recordSet=None, width=(None, '%'), height=(None, "px"), align='center', helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.vignets.text({"title": "New Python Framework", 'value': "A new Python Web Framework", 'color': 'green',
                            'icon': 'fab fa-python', 'colorTitle': 'darkgreen'})

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextComp.TextWithBorder`

    Attributes:
    ----------
    :param recordSet:
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param align:
    :param helper: Optional. A tooltip helper
    :param profile: Optional. A flag to set the component performance storage
    """
    html_text = html.HtmlTextComp.TextWithBorder(self.context.rptObj, recordSet, width, height, align, helper, options, profile)
    return html_text

  def bars(self):
    pass

  def line(self):
    pass




from epyk.core.data.DataClass import DataClass


class OptionConfig(DataClass):

  @property
  def responsive(self):
    """
    Description:
    ------------

    Related Pages:

      https://plot.ly/javascript/configuration-options/
    """
    return self.get(True)

  @responsive.setter
  def responsive(self, val):
    self.set(val)

  @property
  def editable(self):
    """
    Description:
    ------------

    Related Pages:

      https://plot.ly/javascript/configuration-options/
    """
    return self.get(False)

  @editable.setter
  def editable(self, val):
    self.set(val)

  @property
  def staticPlot(self):
    """
    Description:
    ------------

    Related Pages:

      https://plot.ly/javascript/configuration-options/
    """
    return self.get(None)

  @staticPlot.setter
  def staticPlot(self, val):
    self.set(val)

  @property
  def scrollZoom(self):
    """
    Description:
    ------------

    Related Pages:

      https://plot.ly/javascript/configuration-options/
    """
    return self.get(None)

  @scrollZoom.setter
  def scrollZoom(self, val):
    self.set(val)

  @property
  def managed(self):
    """
    Description:
    ------------
    """
    return self.get(True)

  @managed.setter
  def managed(self, bool):
    self.set(bool)




from epyk.core.css import Defaults_css
from epyk.core.css.styles.attributes import Attrs


class AttrHelp(Attrs):
  def __init__(self, rptObj):
    super(AttrHelp, self).__init__(rptObj)
    self.font_size = Defaults_css.font()
    self.cursor = "pointer"
    self.float = "right"
    self.margin = "1px 4px"




from epyk.core import html


class Sliders(object):
  """
  Description:
  ------------
  This module is relying on some Jquery IU components

  The slider and progress bar components can be fully described on the corresponding website

    - https://jqueryui.com/progressbar/
    - https://jqueryui.com/slider/

  As this module will return those object, all the properties and changes defined in the documentation can be done.
  """
  def __init__(self, context):
    self.context = context

  def slider(self, number=0, minimum=0, maximum=100, width=(100, '%'), height=(None, 'px'), htmlCode=None, attrs=None,
             helper=None, options=None, profile=None):
    """
    Description:
    ------------
    Add a Jquery UI slider object to the page

    Usage::

      rptObj.ui.slider(40)
      rptObj.ui.slider([1, 2, 3, 4, 5, 6, 7])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Slider`

    Related Pages:

			https://jqueryui.com/slider/

    Attributes:
    ----------
    :param number:
    :param minimum:
    :param maximum:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    if isinstance(number, list):
      minimum, maximum = min(number), max(number)
      number = minimum
    html_slider = html.HtmlEvent.Slider(self.context.rptObj, number, minimum, maximum, width, height,  attrs or {}, helper,
                                        options or {}, htmlCode, profile)
    return html_slider

  def date(self, value, min=None, max=None, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
             helper=None, options=None, profile=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.SliderDate`

    :param value:
    :param min:
    :param max:
    :param width:
    :param height:
    :param htmlCode:
    :param attrs:
    :param helper:
    :param options:
    :param profile:
    """
    options = options or {}
    html_slider = html.HtmlEvent.SliderDate(self.context.rptObj, value, min, max, width, height, attrs or {}, helper,
                                            options or {}, htmlCode, profile)
    return html_slider

  def date_range(self, value1, value2, min=None, max=None, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
                 helper=None, options=None, profile=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.SliderDate`

    :param value1:
    :param value2:
    :param min:
    :param max:
    :param width:
    :param height:
    :param htmlCode:
    :param attrs:
    :param helper:
    :param options:
    :param profile:
    """
    options = options or {}
    options['range'] = True
    html_slider = html.HtmlEvent.SliderDates(self.context.rptObj, [value1, value2], min, max, width, height, attrs or {}, helper,
                                             options or {}, htmlCode, profile)
    return html_slider

  def range(self, values, min=0, max=100, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
             helper=None, options=None, profile=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Range`

    :param values:
    :param min:
    :param max:
    :param width:
    :param height:
    :param htmlCode:
    :param attrs:
    :param helper:
    :param options:
    :param profile:
    """
    options = options or {}
    options['range'] = True
    html_slider = html.HtmlEvent.Range(self.context.rptObj, values, min, max, width, height, attrs or {}, helper,
                                         options or {}, htmlCode, profile)
    return html_slider

  def progressbar(self, number=0, total=100, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
                  helper=None, options=None, profile=None):
    """
    Description:
    ------------
    Add a progress bar component to the page

    Usage::

      rptObj.ui.sliders.progressbar(300)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.ProgressBar`

    Related Pages:

			https://jqueryui.com/progressbar/

    Attributes:
    ----------
    :param number: A number (by default between 0 and 100)
    :param total: A number
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param htmlCode:
    :param attrs:
    :param helper:
    :param profile:
    """
    html_pr = html.HtmlEvent.ProgressBar(self.context.rptObj, number, total, width, height,  attrs or {}, helper,
                                         options or {}, htmlCode, profile)
    return html_pr

  def progress(self, number=0, total=100, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
                  helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param number:
    :param total:
    :param width:
    :param height:
    :param htmlCode:
    :param attrs:
    :param helper:
    :param options:
    :param profile:
    """
    progress = self.progressbar(number, total, width, height, htmlCode, attrs, helper, options, profile)
    progress.style.css.border_radius = '50px'
    progress.style.css.background = self.context.rptObj.theme.greys[4]
    progress.options.css({"border-radius": "50px", "border-color": self.context.rptObj.theme.greys[5]})
    return progress




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsPackage, DataAttrs


class VisDataSet(JsPackage):
  lib_alias = {'css': 'vis', 'js': 'vis'}

  def __init__(self, src, varName, data, setVar=True):
    super(VisDataSet, self).__init__(src=src, varName=varName, selector="new vis.DataSet(%s)" % data, setVar=setVar)

  @property
  def length(self):
    """
    Description:
    -----------
    The number of items in the DataSet.
    """
    return JsObjects.JsNumber.JsNumber("%s.length" % self.getStr())

  def add(self, jsData):
    """
    Description:
    -----------
    Add one or multiple items to the DataSet. data can be a single item or an array with items.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param jsData: data can be a single item or an array with items.

    :return: The function returns an array with the ids of the added items. See section Data Manipulation.
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsArray.JsArray("%s.add(%s)" % (self.varId, jsData))

  def clear(self, senderId=None):
    """
    Description:
    -----------
    Clear all data from the DataSet

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param senderId: The record ID to be removed

    :return: The function returns an array with the ids of the removed items.
    """
    if senderId is None:
      return JsObjects.JsArray.JsArray("%s.clear()" % (self.varId))

    senderId = JsUtils.jsConvertData(senderId, None)
    return JsObjects.JsArray.JsArray("%s.clear(%s)" % (self.varId, senderId))

  def distinct(self, field):
    """
    Description:
    -----------
    Find all distinct values of a specified field

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param field:

    :return: Returns an unordered array containing all distinct values. If data items do not contain the specified field are ignored.
    """
    return JsObjects.JsArray.JsArray("%s.distinct(%s)" % (self.varId, JsUtils.jsConvertData(field, None)))

  def flush(self):
    """
    Description:
    -----------
    Flush queued changes. Only available when the DataSet is configured with the option queue, see section Construction.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html
    """
    return JsObjects.JsObject.JsObject("%s.flush()" % self.varId)

  def forEach(self, callback, options=None):
    """
    Description:
    -----------
    Execute a callback function for every item in the dataset.

    Attributes:
    ----------
    :param callback:
    :param options:

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.forEach()" % self.varId)

  def map(self, callback):
    pass

  def max(self, field):
    """
    Description:
    -----------
    Find the item with maximum value of specified field

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param field:

    :return: Returns null if no item is found.
    """
    return JsObjects.JsObject.JsObject("%s.max(%s)" % (self.varId, JsUtils.jsConvertData(field, None)))

  def min(self, field):
    """
    Description:
    -----------
    Find the item with minimum value of specified field

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param field:

    :return: Returns null if no item is found.
    """
    return JsObjects.JsObject.JsObject("%s.min(%s)" % (self.varId, JsUtils.jsConvertData(field, None)))

  def update(self):
    pass

  def on(self, event, callback):
    pass

  def off(self, event, callback):
    pass

  def remove(self, ids, senderId=None):
    """
    Description:
    -----------
    Remove a data item or an array with items

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param ids:
    :param senderId:

    :return:
    """
    ids = JsUtils.jsConvertData(ids, None)
    return JsObjects.JsArray.JsArray("%s.remove(%s)" % (self.varId, ids))

  def setOptions(self, options):
    """
    Description:
    -----------
    Set options for the DataSet.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param options:
    """
    return self.fnc_closure("setOptions(%s)" % options)

  def get(self, i):
    return JsObjects.JsObject.JsObject("%s.get(%s)" % (self.varId, i))

  def getIds(self, options=None):
    """
    Description:
    -----------
    Get ids of all items or of a filtered set of items

    Available options are described in section Data Selection, except that options fields and type are not applicable in case of getIds

    Related Pages:

			https://visjs.github.io/vis-data/data/dataset.html

    Attributes:
    ----------
    :param options:
    """
    if options is not None:
      options = JsUtils.jsConvertData(options, None)
      return JsObjects.JsArray.JsArray("%s.getIds(%s)" % (self.getStr(), options))

    return JsObjects.JsArray.JsArray("%s.getIds()" % self.varId)

  def options(self):
    """
    Description:
    -----------
    Create a new option object on the Python side for DataViz
    """
    return VisDataOptions(self.src)


class VisDataOptions(DataAttrs):

  def align(self, position):
    """
    Description:
    -----------

    Related Pages:
:param position:
    """
    return self.attr("align", JsUtils.jsConvertData(position, None))

  def queue_delay(self, n=None):
    """
    Description:
    -----------
    The queue will be flushed automatically after an inactivity of this delay in milliseconds. Default value is null

    Related Pages:
:param n:
    """
    if n is None:
      n = self._report.js.data.null
    return self.attr("delay", JsUtils.jsConvertData(n, None))

  def queue_max(self, n=None):
    """
    Description:
    -----------
    When the queue exceeds the given maximum number of entries, the queue is flushed automatically. Default value is

    Related Pages:
:param n:
    """
    if n is None:
      n = self._report.js.number.POSITIVE_INFINITY
    return self.attr("max", JsUtils.jsConvertData(n, None))

  def autoResize(self, flag):
    """
    Description:
    -----------

    Related Pages:
:param flag:
    """
    return self.attr("autoResize", JsUtils.jsConvertData(flag, None))

  def clickToUse(self, flag):
    """
    Description:
    -----------

    Related Pages:

			https://visjs.github.io/vis-timeline/examples/timeline/interaction/clickToUse.html

    Related Pages:
:param flag:
    """
    return self.attr("clickToUse", JsUtils.jsConvertData(flag, None))

  def configure(self, flag):
    """
    Description:
    -----------

    Related Pages:
:param flag:
    """
    return self.attr("configure", JsUtils.jsConvertData(flag, None))

  def dataAttributes(self, strings):
    """
    Description:
    -----------

    Related Pages:
:param strings:
    """
    return self.attr("dataAttributes", JsUtils.jsConvertData(strings, None))

  @property
  def editable(self):
    pass

  def end(self, object):
    """
    Description:
    -----------

    Related Pages:
:param object:
    """
    return self.attr("end", JsUtils.jsConvertData(object, None))

  def format(self, object):
    """
    Description:
    -----------

    Related Pages:
:param object:
    """
    return self.attr("format", JsUtils.jsConvertData(object, None))

  @property
  def groupEditable(self):
    pass

  def groupHeightMode(self, text):
    """
    Description:
    -----------

    Related Pages:
:param text:
    """
    return self.attr("groupHeightMode", JsUtils.jsConvertData(text, None))

  def groupOrder(self, text):
    """
    Description:
    -----------

    Related Pages:
:param text:
    """
    return self.attr("groupOrder", JsUtils.jsConvertData(text, None))

  def groupOrderSwap(self, fnc):
    """
    Description:
    -----------

    Related Pages:
:param fnc:
    """

  def groupTemplate(self, fnc):
    """
    Description:
    -----------

    Related Pages:
:param fnc:
    """

  def height(self, n):
    """
    Description:
    -----------

    Related Pages:
:param n:
    """
    return self.attr("height", JsUtils.jsConvertData(n, None))

  def hiddenDates(self, object):
    """
    Description:
    -----------

    Related Pages:
:param object:
    """

  def horizontalScroll(self, flag):
    """
    Description:
    -----------

    Related Pages:
:param flag:
    """
    return self.attr("horizontalScroll", JsUtils.jsConvertData(flag, None))


class VisDataView(JsPackage):
  lib_alias = {'css': 'vis', 'js': 'vis'}

  def __init__(self, src, varName, data, setVar=True):
    super(VisDataView, self).__init__(src=src, varName=varName, selector="new vis.DataView(%s)" % data, setVar=setVar)

  @property
  def length(self):
    """
    Description:
    -----------
    The number of items in the DataSet.
    """
    return JsObjects.JsNumber.JsNumber("%s.length" % self.varId)

  def get(self, options=None, data=None):
    """
    Description:
    -----------
    Get a single item, multiple items, or all items from the DataView.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html

		Attributes:
    ----------
    :param options:
    :param data:
    """
    if data is None:
      if options is None:
        return JsObjects.JsObject.JsObject("%s.get()" % (self.getStr()))

      options = JsUtils.jsConvertData(options, None)
      return JsObjects.JsObject.JsObject("%s.get(%s)" % (self.getStr(), options))

    options = JsUtils.jsConvertData(options, None)
    data = JsUtils.jsConvertData(data, None)
    return JsObjects.JsObject.JsObject("%s.get(%s, %s)" % (self.varId, options, data))

  def getByIds(self, ids, options=None, data=None):
    """
    Description:
    -----------
    Get a single item, multiple items, or all items from the DataView.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html

    Attributes:
    ----------
    :param ids:
    :param options:
    :param data:
    """
    ids = JsUtils.jsConvertData(ids, None)
    if data is None:
      if options is None:
        return JsObjects.JsObject.JsObject.get("%s.get(%s)" % (self.varId, ids))

      options = JsUtils.jsConvertData(options, None)
      return JsObjects.JsObject.JsObject.get("%s.get(%s, %s)" % (self.varId, ids, options))

    options = JsUtils.jsConvertData(options, None)
    data = JsUtils.jsConvertData(data, None)
    return JsObjects.JsObject.JsObject.get("%s.get(%s, %s, %s)" % (self.varId, ids, options, data))

  def getDataSet(self, vanName):
    """
    Description:
    -----------
    Get the DataSet to which the DataView is connected.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html

		Attributes:
    ----------
    :param vanName:
    """
    return VisDataSet(src=self.src, data="%s.getDataSet()" % self.varId, varName=vanName)

  def getIds(self, options=None):
    """
    Description:
    -----------
    Get ids of all items or of a filtered set of items.
    Available options are described in section Data Selection, except that options fields and type are not applicable in case of getIds.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html

    Attributes:
    ----------
    :param options:
    """
    if options is not None:
      options = JsUtils.jsConvertData(options, None)
      return JsObjects.JsArray.JsArray("%s.getIds(%s)" % (self.varId, options))

    return JsObjects.JsArray.JsArray("%s.getIds()" % self.varId)

  def off(self, event, callback):
    """
    Description:
    -----------
    Unsubscribe from an event, remove an event listener.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html
    https://visjs.github.io/vis-data/data/dataview.html#Subscriptions

    Attributes:
    ----------
    :param event:
    :param callback:
    """

  def on(self, event, callback):
    """
    Description:
    -----------
    Subscribe to an event, add an event listener.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html
    https://visjs.github.io/vis-data/data/dataview.html#Subscriptions

    Attributes:
    ----------
    :param event:
    :param callback:
    """

  def refresh(self):
    """
    Description:
    -----------
    Refresh the filter results of a DataView.

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html
    """
    return self.fnc_closure("refresh()")

  def setData(self, data):
    """
    Description:
    -----------
    Replace the DataSet of the DataView. Parameter data can be a DataSet or a DataView.

    Attributes:
    ----------
    :param data:
    """
    return self.fnc("setData(%s)" % JsUtils.jsConvertData(data, None))

  def options(self):
    """
    Description:
    -----------
    Create a new option object on the Python side for DataViz
    """
    return VisDataOptions(self.src)


class VisDataGroups(DataAttrs):

  def className(self, value):
    """
    Description:
    -----------
    This field is optional. A className can be used to give groups an individual css style. For example, when a group has className 'red', one can define a css style .red { color: red; } .

    Related Pages:

			https://visjs.github.io/vis-timeline/docs/timeline/

    Attributes:
    ----------
    :param value:
    """
    return self.attr("className", JsUtils.jsConvertData(value, None))

  @property
  def content(self):
    """
    Description:
    -----------
    The contents of the group. This can be plain text, html code or an html element.

    Attributes:
    ----------
    :param item:
    """
    return self._attrs["content"]

  @content.setter
  def content(self, item):
    return self.attr("content", item)

  def options(self):
    pass


class VisGroups(JsPackage):
  lib_selector = "new vis.DataSet()"

  def add_group(self):
    return VisDataGroups(self.src)

  def add(self, groups):
    for name in groups:
      grp = self.add_group()
      grp.content = name
      self.fnc_closure("add(%s)" % grp.toStr())
    return self


class VisNetworkNode(JsPackage):
  pass


class VisNetworkEdge(JsPackage):
  def getPositions(self, nodeIds):
    """
    Description:
    -----------
    Returns the x y positions in canvas space of the nodes with the supplied nodeIds as an object

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/

    Attributes:
    ----------
    :param nodeIds:
    """

  def storePositions(self):
    """
    Description:
    -----------
    When using the vis.DataSet to load your nodes into the network, this method will put the X and Y positions of all nodes into that dataset.

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/
    """

  def moveNode(self, nodeId, x, y):
    """
    Description:
    -----------
    You can use this to programatically move a node. The supplied x and y positions have to be in canvas space!

    Attributes:
    ----------
    :param nodeId:
    :param x:
    :param y:
    """

  def getBoundingBox(self, nodeId):
    """
    Description:
    -----------
    Returns a bounding box for the node including label in the format:
    {top: Number, left: Number, right: Number, bottom: Number}

    Attributes:
    ----------
    :param nodeId:
    """

  def getConnectedNodes(self, nodeId, direction=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param nodeId:
    :param direction:
    """

  def getConnectedEdges(self, nodeId):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param nodeId:
    """


class VisNetwork(JsPackage):
  lib_alias = {'css': 'vis', 'js': 'vis'}

  def destroy(self):
    """
    Description:
    -----------
    Remove the network from the DOM and remove all Hammer bindings and references.

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/
    """
    return JsObjects.JsVoid("%s.destroy()" % self.varId)

  def setData(self, data):
    """
    Description:
    -----------
    Override all the data in the network.
    If stabilization is enabled in the physics module, the network will stabilize again.
    This method is also performed when first initializing the network.

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/

    Attributes:
    ----------
    :param data:
    """
    return JsObjects.JsVoid("%s.setData(%s)" % (self.varId, JsUtils.jsConvertData(data, None)))

  def setOptions(self, options):
    """
    Description:
    -----------
    Set the options. All available options can be found in the modules above.
    Each module requires it's own container with the module name to contain its options.

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/

    Attributes:
    ----------
    :param options:
    """
    return JsObjects.JsVoid("%s.setOptions(%s)" % (self.varId, JsUtils.jsConvertData(options, None)))

  def redraw(self):
    """
    Description:
    -----------
    Redraw the network.

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/
    """
    return JsObjects.JsVoid("%s.redraw()" % self.varId)

  def setSize(self, width, height):
    """
    Description:
    -----------
    Set the size of the canvas. This is automatically done on a window resize.

    Related Pages:

			https://visjs.github.io/vis-network/docs/network/

    Attributes:
    ----------
    :param width: Integer. The width
    :param height: Integer. The height
    """
    if isinstance(width, int):
      width = "%spx" % width

    if isinstance(height, int):
      height = "%spx" % height
    return JsObjects.JsVoid("%s.setSize(%s, %s)" % (self.varId, JsUtils.jsConvertData(width, None), JsUtils.jsConvertData(height, None)))


class VisTimeline(JsPackage):
  lib_alias = {'css': 'vis', 'js': 'vis'}

  def addItem(self, data):
    """
    Description:
    -----------#

    Attributes:
    ----------
    :param data:
    """
    return JsObjects.JsVoid("%s.itemsData.add(%s)" % (self.varId, JsUtils.jsConvertData(data, None)))
    #return JsObjects.JsVoid("console.log(%s)" % (self.varId))#, JsUtils.jsConvertData(data, None)))

  def addItems(self, data):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    """
    return JsObjects.JsVoid("%s.forEach(function(data) {%s.itemsData.add(data)})" % (JsUtils.jsConvertData(data, None), self.varId))

  def setItems(self, data):
    """
    Description:
    -----------
    Replace the DataSet of the DataView. Parameter data can be a DataSet or a DataView.

    Attributes:
    ----------
    :param data:
    """
    return JsObjects.JsVoid("%s.setItems(%s)" % (self.varId, JsUtils.jsConvertData(data, None)))

  def setGroups(self, groups):
    """
    Description:
    -----------
    Set a data set with groups for the Graph2d. groups can be an Array with Objects, a DataSet, or a DataView.
    For each of the groups, the items of the Graph2d are filtered on the property group, which must correspond with the id of the group.

    https://visjs.github.io/vis-timeline/docs/graph2d/

    Attributes:
    ----------
    :param groups:
    """
    return JsObjects.JsVoid("%s.setGroups(%s)" % (self.varId, JsUtils.jsConvertData(groups, None)))

  def setOptions(self, options):
    """
    Description:
    -----------
    Set or update options. It is possible to change any option of the Graph2d at any time.
    You can for example switch orientation on the fly.

    https://visjs.github.io/vis-timeline/docs/graph2d/

    Attributes:
    ----------
    :param options:
    """
    return JsObjects.JsVoid("%s.setOptions(%s)" % (self.varId, JsUtils.jsConvertData(options, None)))

  def destroy(self):
    """
    Description:
    -----------
    Destroy the Graph2d. The Graph2d is removed from memory. all DOM elements and event listeners are cleaned up.

    Related Pages:

      https://visjs.github.io/vis-network/docs/network/
    """
    return JsObjects.JsVoid("%s.destroy()" % self.varId)

  def redraw(self):
    """
    Description:
    -----------
    Force a redraw of the Graph2d. Can be useful to manually redraw when option autoResize=false.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/
    """
    return JsObjects.JsVoid("%s.redraw()" % self.varId)

  def fit(self):
    """
    Description:
    -----------
    Adjust the visible window such that it fits all items.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/
    """
    return JsObjects.JsVoid("%s.fit()" % self.varId)


class VisGraph3D(JsPackage):
  lib_alias = {'css': 'vis', 'js': 'vis'}

  def animationStart(self):
    """
    Description:
    -----------
    Start playing the animation.
    Only applicable when animation data is available.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/
    """
    return self.fnc_closure("animationStart()")

  def animationStop(self):
    """
    Description:
    -----------
    Stop playing the animation.
    Only applicable when animation data is available.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/
    """
    return self.fnc_closure("animationStop()")

  def getCameraPosition(self):
    """
    Description:
    -----------
    Returns an object with parameters horizontal, vertical and distance, which each one of them is a number, representing the rotation and position of the camera.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/
    """
    return "getCameraPosition()"

  def redraw(self):
    """
    Description:
    -----------
    Redraw the graph. Useful after the camera position is changed externally, when data is changed, or when the layout of the webpage changed.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/
    """
    return JsObjects.JsVoid("%s.redraw()" % self.varName)

  def setData(self, data):
    """
    Description:
    -----------
    Replace the data in the Graph3d.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/

    Attributes:
    ----------
    :param data:
    """
    return self.fnc_closure("setData(%s)" % JsUtils.jsConvertData(data, None))

  def setOptions(self, options):
    """
    Description:
    -----------
    Update options of Graph3d. The provided options will be merged with current options.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/

    Attributes:
    ----------
    :param options:
    """
    return self.fnc_closure("setOptions(%s)" % JsUtils.jsConvertData(options, None))

  def setSize(self, width, height):
    """
    Description:
    -----------
    Parameters width and height are strings, containing a new size for the graph. Size can be provided in pixels or in percentages.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/

    Attributes:
    ----------
    :param width:
    :param height:
    """
    width = JsUtils.jsConvertData(width, None)
    height = JsUtils.jsConvertData(height, None)
    return self.fnc_closure("setSize(%s, %s)" % (width, height))

  def setCameraPosition(self, pos):
    """
    Description:
    -----------
    Set the rotation and position of the camera.
    Parameter pos is an object which contains three parameters: horizontal, vertical, and distance.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/

    Attributes:
    ----------
    :param pos:
    """
    pos = JsUtils.jsConvertData(pos)
    return self.fnc_closure("setCameraPosition(%s)" % pos)

  def onCameraPositionChange(self, jsFnc):
    """
    Description:
    -----------
    The camera position changed. Fired after the user modified the camera position by moving (dragging) the graph, or by zooming (scrolling), but not after a call to setCameraPosition method.

    Related Pages:

			https://visjs.github.io/vis-graph3d/docs/graph3d/

    Attributes:
    ----------
    :param jsFnc:
    """
    return self.fnc_closure("on('cameraPositionChange', %s)" % jsFnc)


class VisGraph2D(JsPackage):
  lib_alias = {'css': 'vis', 'js': 'vis'}

  def destroy(self):
    """
    Description:
    -----------
    Destroy the Graph2d. The Graph2d is removed from memory. all DOM elements and event listeners are cleaned up.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/
    """
    return JsObjects.JsVoid("%s.destroy()" % self.varId)

  def fit(self):
    """
    Description:
    -----------
    Adjust the visible window such that it fits all items.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/
    """
    return JsObjects.JsVoid("%s.fit()" % self.varId)

  def getCurrentTime(self):
    """
    Description:
    -----------
    Get the current time. Only applicable when option showCurrentTime is true.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/
    """
    return JsObjects.JsVoid("%s.getCurrentTime()" % self.varId)

  def getDataRange(self):
    """
    Description:
    -----------
    Get the range of all the items as an object containing min: Date and max: Date.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/
    """
    return JsObjects.JsVoid("%s.getCurrentTime()" % self.varId)

  def redraw(self):
    """
    Description:
    -----------
    Redraw the graph. Useful after the camera position is changed externally, when data is changed, or when the layout of the webpage changed.

    Related Pages:

      https://visjs.github.io/vis-graph3d/docs/graph3d/
    """
    return JsObjects.JsVoid("%s.redraw()" % self.varId)

  def setItems(self, items):
    """
    Description:
    -----------
    Set a data set with items for the Graph2d. items can be an Array with Objects, a DataSet, or a DataView.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/

    Attributes:
    ----------
    :param items:
    """
    items = JsUtils.jsConvertData(items, None)
    return JsObjects.JsVoid("%s.setItems(%s)" % (self.varId, items))

  def setOptions(self, options):
    """
    Description:
    -----------
    Set or update options. It is possible to change any option of the Graph2d at any time.
    You can for example switch orientation on the fly.

    Related Pages:

      https://visjs.github.io/vis-timeline/docs/graph2d/

    Attributes:
    ----------
    :param options:
    """
    options = JsUtils.jsConvertData(options, None)
    return JsObjects.JsVoid("%s.setOptions(%s)" % (self.varId, options))




from epyk.core.data.DataClass import DataClass


class InnerButtons(DataClass):

  @property
  def text(self):
    """
    Description:
    -----------
    Being able to let your users know what will happen when they activate a button is obviously fundamentally important to the Buttons extension and this option provides exactly that ability.
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.text
    """
    return self._attrs['text']

  @text.setter
  def text(self, val):
    self._attrs['text'] = val

  @property
  def action(self):
    """
    Description:
    -----------
    This function defined the action that the button will take when activated by the end user.
    This will normally be to perform some operation on the DataTable, but can be absolutely anything since the function can be defined by yourself.
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.action
    """
    return self._attrs['action']

  @action.setter
  def action(self, val):
    self._attrs['action'] = val

  @property
  def attr(self):
    """
    Description:
    -----------
    This option provides the ability to set any arbitrary attribute on the button's HTML element in the document.
    This can be useful for customisation of attributes such as ARIA assistive attributes, setting a custom id for selection / styling, custom data parameters, etc.
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.attr
    """
    return self._attrs['attr']

  @attr.setter
  def attr(self, val):
    self._attrs['attr'] = val

  @property
  def className(self):
    """
    Description:
    -----------
    Set the class name for the button.
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.className
    """
    return self._attrs['className']

  @className.setter
  def className(self, val):
    self._attrs['className'] = val

  @property
  def available(self):
    """
    Description:
    -----------
    Ensure that any requirements have been satisfied before initialising a button
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.available
    """
    return self._attrs['available']

  @available.setter
  def available(self, val):
    self._attrs['available'] = val

  @property
  def enabled(self):
    """
    Description:
    -----------
    This option provides the ability to set the initial enabled state of a button using a boolean value.
    It is most likely to be of use when using the API methods that can control the button's enabled state after initialisation:
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.enabled
    """
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, val):
    self._attrs["enabled"] = val

  @property
  def extend(self):
    """
    Description:
    -----------
    Define which button type the button should be based on.

    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.extend
    """
    return self._attrs["extend"]

  @extend.setter
  def extend(self, val):
    self._attrs["extend"] = val

  @property
  def key(self):
    """
    Description:
    -----------
    Define an activation key for a button.

    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.key
    """
    return self._attrs["key"]

  @key.setter
  def key(self, val):
    self._attrs["key"] = val

  @property
  def name(self):
    """
    Description:
    -----------
    Define a name for a button.

    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.name
    """
    return self._attrs["name"]

  @name.setter
  def name(self, val):
    self._attrs["name"] = val

  @property
  def tag(self):
    """
    Description:
    -----------
    Define a tag  for a button.

    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.tag
    """
    return self._attrs["tag"]

  @tag.setter
  def tag(self, val):
    self._attrs["tag"] = val

  @property
  def titleAttr(self):
    """
    Description:
    -----------
    This option provides the ability to set the title attribute for the button.
    This can be used to provide a detailed description of the button, or a simple text name if you choose to display an icon in the button rather than text (using the buttons.buttons.text option).

    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.titleAttr
    """
    return self._attrs["titleAttr"]

  @titleAttr.setter
  def titleAttr(self, val):
    self._attrs["titleAttr"] = val

class Buttons(DataClass):

  class Dom(DataClass):

    class Container(DataClass):
      @property
      def tag(self):
        return self._attrs['tag']

      @tag.setter
      def tag(self, val):
        self._attrs['tag'] = val

      @property
      def className(self):
        return self._attrs['className']

      @className.setter
      def className(self, val):
        self._attrs['className'] = val


    class ButtonLiner(DataClass):
      @property
      def tag(self):
        return self._attrs['tag']

      @tag.setter
      def tag(self, val):
        self._attrs['tag'] = val

      @property
      def className(self):
        return self._attrs['className']

      @className.setter
      def className(self, val):
        self._attrs['className'] = val


    class Collection(DataClass):
      @property
      def tag(self):
        return self._attrs['tag']

      @tag.setter
      def tag(self, val):
        self._attrs['tag'] = val

      @property
      def className(self):
        return self._attrs['className']

      @className.setter
      def className(self, val):
        self._attrs['className'] = val


    class DomButton(DataClass):
      @property
      def tag(self):
        return self._attrs['tag']

      @tag.setter
      def tag(self, val):
        self._attrs['tag'] = val

      @property
      def className(self):
        return self._attrs['className']

      @className.setter
      def className(self, val):
        self._attrs['className'] = val

      @property
      def disabled(self):
        return self._attrs['disabled']

      @disabled.setter
      def disabled(self, val):
        self._attrs['disabled'] = val

      @property
      def active(self):
        return self._attrs['active']

      @active.setter
      def active(self, val):
        self._attrs['active'] = val


    @property
    def buttonContainer(self):
      """
      Description:
      -----------
      This option provides the ability for each button to be wrapped in another element.
      This can be useful for cases where the styling framework requires a list of buttons,
      where the list element is just a container and not the button itself (Zurb Foundation requires this for example).
      Related Pages:

			https://datatables.net/reference/option/buttons.buttons.buttonContainer
      """
      return self.sub_data('buttonContainer', self.Container)

    @property
    def buttonLiner(self):
      """
      Description:
      -----------
      This option controls the HTML tag that is used as the liner for each button.
      This can be particularly useful for adding complex styling rules to buttons.
      It can also be disabled if you wish to have minimal markup in your document.
      Related Pages:

			https://datatables.net/reference/option/buttons.buttons.buttonLiner
      """
      return self.sub_data('buttonLiner', self.ButtonLiner)

    @property
    def collection(self):
      """
      Description:
      -----------
      This option controls the HTML tag that is used for the element that is displayed when a collection button is triggered.
      It in turn contains the sub-buttons of the collection.
      Related Pages:

			https://datatables.net/reference/option/buttons.buttons.collection
      """
      return self.sub_data('collection', self.Collection)

    @property
    def button(self):
      """
      Description:
      -----------
      This option controls the HTML tag that is used to create each individual button.
      With this option the tag type and class name can be specified using the tag and className properties of this object.
      Related Pages:

			https://datatables.net/reference/option/buttons.buttons.button
      """
      return self.sub_data('button', self.DomButton)


  @property
  def enabled(self):
    """
    Description:
    -----------
    This option provides the ability to set the initial enabled state of a button using a boolean value.
    It is most likely to be of use when using the API methods that can control the button's enabled state after initialisation:
    Related Pages:

			https://datatables.net/reference/option/buttons.buttons.enabled
    """
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, val):
    self._attrs["enabled"] = val

  @property
  def buttons(self):
    return self.sub_data_enum('buttons', InnerButtons)

  @property
  def dom(self):
    return self.sub_data('dom', self.Dom)

  @property
  def name(self):
    """
    Description:
    -----------
    Set a name for the instance for the group selector.
    Related Pages:

			https://datatables.net/reference/option/buttons.name
    """
    return self._attrs['name']

  @name.setter
  def name(self, val):
    self._attrs['name'] = val



from epyk.core.py import PyDates

print(PyDates.PyDates().today)
print(PyDates.PyDates().now)
print(PyDates.PyDates().cob)
print(PyDates.PyDates().month_end)
print(PyDates.PyDates().months)
print(PyDates.PyDates().date_from_alias("T", "2019-08-08"))



from epyk.core.html import geo


class GeoGoogle(object):

  def __init__(self, context):
    self.parent = context
    self.chartFamily = "GoogleMaps"

  def maps(self, lattitude, longitude, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    :param lattitude:
    :param longitude:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    dflt_options = {"center": (lattitude, longitude), 'mapTypeId': 'ROADMAP', 'zoom': 10}
    if options is not None:
      dflt_options.update(options)
    geo_chart = geo.GeoGoogle.ChartGeoGoogle(self.parent.context.rptObj, width, height, dflt_options, htmlCode, profile)

    return geo_chart

  def satellite(self, lattitude, longitude, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    :param lattitude:
    :param longitude:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    dflt_options = {"center": (lattitude, longitude), 'mapTypeId': 'satellite', 'zoom': 10}
    if options is not None:
      dflt_options.update(options)
    geo_chart = geo.GeoGoogle.ChartGeoGoogle(self.parent.context.rptObj, width, height, dflt_options, htmlCode, profile)
    return geo_chart

  def terrain(self, lattitude, longitude, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    :param lattitude:
    :param longitude:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    dflt_options = {"center": (lattitude, longitude), 'mapTypeId': 'terrain', 'zoom': 10}
    if options is not None:
      dflt_options.update(options)
    geo_chart = geo.GeoGoogle.ChartGeoGoogle(self.parent.context.rptObj, width, height, dflt_options, htmlCode, profile)
    return geo_chart

  def current(self, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    dflt_options = {"center": "navigator.geolocation.getCurrentPosition(function(position){console.log(position.coords.latitude); return (position.coords.latitude, position.coords.longitude)})", 'mapTypeId': 'ROADMAP', 'zoom': 10}
    if options is not None:
      dflt_options.update(options)
    geo_chart = geo.GeoGoogle.ChartGeoGoogle(self.parent.context.rptObj, width, height, dflt_options, htmlCode, profile)
    return geo_chart

  def streetview(self):
    pass
